;-----------------------------------------------------------
; Wacom serial tablet and bus mouse driver WMDRV
; version 0.5a
; wmdrv.asm
; Copyrigt (C) 2022-2025 あめさろだ (rainy-srd, r-srd)
; Released under the 3-clause BSD License.
;-----------------------------------------------------------

; このファイルで定義されている関数は特記がない限り以下の仕様
; * ax レジスタ以外は破壊しない
; * 32bit レジスタの上位 16bit は破壊されうる
; * メインルーチン及びその jmp 先のルーチンは任意の汎用レジスタを破壊しうる

.8086

CODE SEGMENT
	assume	cs:CODE, ds:CODE, ss:CODE, es:CODE
	org	100h

;-----------------------------------------------------------
; 定数等
;---------------------------------------
DISPLAY_WIDTH	equ 640
DISPLAY_HEIGHT	equ 400

CTRL_VECTOR	equ 33h		; ドライバ制御ベクタ

; out 5fh, al して待つマクロ
WAIT_5FH MACRO count:REQ
	REPEAT count
		out	5fh, al
	ENDM
ENDM

; シフトマクロ
SHFT_L MACRO reg:REQ, count:REQ
	REPEAT count
		shl	reg, 1
	ENDM
ENDM
SHFT_R MACRO reg:REQ, count:REQ
	REPEAT count
		shr	reg, 1
	ENDM
ENDM

; プログラム実行時のエラー
PROG_NO_ERR		equ 00h
PROG_ERR_ARG		equ 01h ; 引数のパース失敗
PROG_ERR_TABLET_SETUP	equ 02h ; タブレットセットアップ時のエラー
PROG_ERR_ALREADY_STAYED	equ 03h ; 既に常駐している
PROG_ERR_NOT_STAYED	equ 04h ; 常駐していない
PROG_ERR_INVALID_MARGIN	equ 05h ; 余白幅の値がおかしい
PROG_ERR_NO_2NDCCU	equ 06h ; 2ndCCU が見つからない
PROG_ERR_NO_9861K	equ 07h ; PC-9861K 互換の RS-232C が見つからない
PROG_ERR_NO_RSA_98II	equ 08h ; RSA-98II, III が見つからない

; プログラム動作関係フラグ
PROG_FLAG_DO_STAY	equ 01h ; 常駐して終了するか?

; プログラム動作モード
PROG_MODE_NO_OPERATE	equ 00h
PROG_MODE_STAY		equ 01h
PROG_MODE_REMOVE	equ 02h
PROG_MODE_HELP		equ 03h


; RS-232C 受信時のエラー
SER_NO_ERR		equ 00h
SER_ERR_PE		equ 04h ; parity error (8251 status)
SER_ERR_OVE		equ 10h ; overrun error (8251 status)
SER_ERR_FE		equ 20h ; framing error (8251 status)
SER_ERR_TSR_OVE		equ 40h ; overrun error (TSR)
SER_ERR_BUF_FULL	equ 80h ; save_tilde_to_cr の読み出し先が溢れた

; RS-232C フラグ
SER_FLAG_PROCESSING	equ 01h ; 受信データ処理中フラグ
SER_FLAG_READING	equ 02h ; データ読み取り中フラグ
SER_FLAG_READ_OK	equ 04h ; 読み取り完了フラグ

; RS-232C 受信処理の動作モード
SER_MODE_NO_OPERATE	equ 0	; 何もしない(データを読んでも捨てる)
SER_MODE_GET_MODEL	equ 1	; モデル名取得時
SER_MODE_GET_AREA	equ 2	; タブレット座標範囲取得時
SER_MODE_PARSE_DATA	equ 3	; タブレットデータパース中

; マウス関係 IO ポート
MOUSE_IO_PORT_A		equ 7fd9h
MOUSE_IO_PORT_B		equ 7fdbh
MOUSE_IO_PORT_C		equ 7fddh
MOUSE_IO_CTRL_REG	equ 7fdfh
MOUSE_IO_CTRL_VAL	equ 93h
MOUSE_IO_INT_FREQ	equ 0bfdbh

; マウスフラグ
MOUSE_FLAG_PROCESSING	equ 01h		; 割込処理中フラグ

; マウスボタン定数
MOUSE_BUTTON_0		equ 01h
MOUSE_BUTTON_1		equ 02h
;MOUSE_BUTTON_2		equ 04h		; 中ボタンは未サポート

; ミッキー/ドット比初期値
MKY_DOT_RATE_INIT	equ 8

; ドライバ制御フラグ(int 33h(ax=00h)で初期化, 動作中に変化する動的な情報)
DRIVER_FLAG_USER_CURSOR	equ 01h ; ユーザ指定のカーソルを使った描画
DRIVER_FLAG_FIRST_DRAW	equ 02h ; 初回のカーソル描画
DRIVER_FLAG_DRAWING	equ 04h ; カーソル描画処理中
DRIVER_FLAG_ERASING	equ 08h ; カーソル消去処理中
DRIVER_FLAG_UPD_TABLET	equ 10h ; タブレットのボタン及び座標情報を更新した
DRIVER_FLAG_UPD_MOUSE	equ 20h ; マウスのボタン及び座標情報を更新した
DRIVER_FLAG_REQ_REDRAW	equ 40h ; 次のマウス割込で必ず再描画する
DRIVER_FLAG_INIT_VAL	equ 02h
; ドライバ制御フラグ(int 33h(ax=00h)で初期化されない)
DRIVER_FLAG_S_SYNC_CURSOR equ 01h ; タブレットとマウスのカーソル位置を同期
DRIVER_FLAG_S_EN_DBG_FUNC equ 02h ; デバッグ用独自ファンクション有効
DRIVER_FLAG_S_INIT_VAL    equ 01h

;通信回線
COM_LINE_1st		equ 00h ; 標準 RS-232C インターフェース
COM_LINE_9861K_2nd	equ 01h ; PC-9861K 第2回線
COM_LINE_9861K_3rd	equ 02h ; PC-9861K 第3回線
COM_LINE_2ndCCU		equ 03h ; 2ndCCU
COM_LINE_RSA_98II	equ 04h ; RSA-98II の 16550 互換モード
COM_LINE_MAX		equ 04h


; ユーザ定義サブルーチン呼び出しフラグ
USER_SUBROUTINE_POS_CHANGE	equ 01h
USER_SUBROUTINE_L_CLICKED	equ 02h
USER_SUBROUTINE_L_RELEASED	equ 04h
USER_SUBROUTINE_R_CLICKED	equ 08h
USER_SUBROUTINE_R_RELEASED	equ 10h
;USER_SUBROUTINE_C_CLICKED	equ 20h		; 中ボタンは未サポート
;USER_SUBROUTINE_C_RELEASED	equ 40h

; ボタン押下及び離し情報を保持する構造体
ButtonInfo STRUC
	click_cnt	dw ?
	release_cnt	dw ?
	last_clk_x	dw ?
	last_clk_y	dw ?
	last_rel_x	dw ?
	last_rel_y	dw ?
ButtonInfo ENDS


; プロトコル関係
PROTO_W_IV		equ 1		; wacom serial protocol
PROTO_W_V		equ 2
PROTO_W_IV_PACKET_LEN	equ 7
PROTO_W_V_PACKET_LEN	equ 9
LONGEST_PACKET_LEN	equ 9

; タブレットの型番
TABLET_MODEL_UD		equ 0
TABLET_MODEL_KT		equ 1
TABLET_MODEL_GD		equ 2
TABLET_MODEL_XD		equ 3

; クリックした扱いにする筆圧のデフォルト値
PRESSURE_THRESHOLD_UD	equ 20
PRESSURE_THRESHOLD_KT	equ 20
PRESSURE_THRESHOLD_GD	equ 100
PRESSURE_THRESHOLD_XD	equ 100

; 上下左右の余白幅のデフォルト値
TABLET_DEFAULT_MARGIN	equ 80


; プログラム引数で指定できるパラメータ
OPT_P_THRESHOLD		equ 01h ; 筆圧閾値(/p)
OPT_TABLET_MARGIN	equ 02h ; 余白幅(/m)
OPT_COM_LINE		equ 04h ; 通信回線(/c)
OPT_IO_ADDR		equ 08h ; IO アドレス(/a)
OPT_INT			equ 10h ; INT (/i)
OPT_MKY			equ 20h ; ミッキー値(/k)
OPT_DEBUG_FUNC		equ 40h ; デバッグ用独自ファンクション有効(/d)
OPT_SYNC_CURSOR		equ 80h ; カーソル同期


;-----------------------------------------------------------
; エントリポイント
;---------------------------------------
start:
	jmp	main

;-----------------------------------------------------------
; 常駐部分
;---------------------------------------
; ドライバ制御割込ハンドラ
ctrl_int_handler PROC
	jmp	near ptr @F
	db 'WMDRV', 0, '0.4a', 0	; 常駐チェックデータ
@@:	;---
	sti		; 他の割込可(各ルーチンで必要に応じて cli/sti する)
	push	ds
	push	si
	push	ax
	;---
	mov	si, cs
	mov	ds, si
	cmp	ax, 0fh
	ja	@F
	add	ax, ax		; word access
	mov	si, ax
	jmp	[tbl_ctrl_proc + si]
@@:	; 独自処理(ax=c0h 以降)
.if(([driver_flags_s] & DRIVER_FLAG_S_EN_DBG_FUNC) \
					&& (ax >= 0c0h) && (ax <= 0c2h))
	sub	ax, 0c0h
	add	ax, ax		; word access
	mov	si, ax
	jmp	[tbl_ctrl_ext_proc + si]
.endif
	;---
ctrl_proc_dummy::
ctrl_proc_end::
	pop	ax
@@:	pop	si
	pop	ds
	;---
	iret
ctrl_proc_end_ax_modify::
	add	sp, 2		; push ax の結果を捨てる
	jmp	@B
ctrl_int_handler ENDP
;--------------------------------------- 00h: 初期化
ctrl_proc_00h PROC
	push	cx
	push	di
	push	es
	cli
	;--- 描画中ならカーソル消去
	cmp	[cursor_counter], 0
	jnz	@F
	dec	[cursor_counter]
	call	erase_cursor
@@:	;--- 0クリア
	cli			; erase_cursor で sti している
	mov	ax, cs
	mov	es, ax
	mov	cx, (offset init_zero_tail - offset init_zero_head) / 2
	xor	ax, ax
	mov	di, offset init_zero_head
	pushf
	cld
	rep	stosw
	popf					; DF 復帰
	pop	es
	pop	di				; cx の pop は後で
	;--- カーソル位置設定
	mov	[cursor_max_x], DISPLAY_WIDTH - 1
	mov	[cursor_max_y], DISPLAY_HEIGHT - 1
	mov	ax, DISPLAY_WIDTH / 2
	mov	bx, DISPLAY_HEIGHT / 2
	mov	[cursor_x], ax
	mov	[cursor_y], bx
	mov	[cursor_tablet_x], ax
	mov	[cursor_tablet_y], bx
	mov	[cursor_mouse_x], ax
	mov	[cursor_mouse_y], bx
	mov	[cursor_0bh_x], ax
	mov	[cursor_0bh_y], bx
	mov	[cursor_user_subr_x], ax
	mov	[cursor_user_subr_y], bx
	mov	[cursor_counter], -1
	;--- 他の設定
	mov	ax, [mky_dot_rate_specify]
	mov	[mky_dot_rate_x], ax
	mov	[mky_dot_rate_y], ax
	mov	[driver_flags], DRIVER_FLAG_INIT_VAL
	;--- マウス割込のマスク解除
	mov	cl, [irq_mouse]
	call	clear_imr_mask
	pop	cx
	;---  マウス割込許可
	push	dx
	mov	dx, MOUSE_IO_PORT_C
	in	al, dx
	and	al, 0efh
	out	dx, al
	pop	dx
	;---
	mov	ax, -1
	mov	bx, 2
	jmp	ctrl_proc_end_ax_modify
ctrl_proc_00h ENDP
;--------------------------------------- 01h: カーソル表示
ctrl_proc_01h PROC
	cli
	; カウンタに1加算して0又は元から0でないなら描画しない
	cmp	[cursor_counter], 0
	jz	@F
	inc	[cursor_counter]
	jnz	Lend
	; 初回なら無条件で描画
@@:	test	[driver_flags], DRIVER_FLAG_FIRST_DRAW
	jnz	draw
	; カーソルが動いていないなら描画しない
	mov	ax, [cursor_last_draw_x]
	mov	si, [cursor_last_draw_y]
	cmp	ax, [cursor_x]
	jnz	draw
	cmp	si, [cursor_y]
	jnz	draw
Lend:	jmp	ctrl_proc_end
draw:	mov	ax, ctrl_proc_end
	push	ax
	jmp	draw_cursor_33h
ctrl_proc_01h ENDP
;--------------------------------------- 02h: カーソル消去
ctrl_proc_02h PROC
	cli
	dec	[cursor_counter]
	cmp	[cursor_counter], -1
	jnz	Lend
	; カーソル消去するのは [cursor_counter] がもともと0だった場合のみ
	push	cx
	push	di
	push	es
	call	erase_cursor
	or	[driver_flags], DRIVER_FLAG_FIRST_DRAW
	pop	es
	pop	di
	pop	cx
Lend:	jmp	ctrl_proc_end
ctrl_proc_02h ENDP
;--------------------------------------- 03h: 状態取得
ctrl_proc_03h PROC
	xor	bh, bh
	cli
	mov	bl, [button_status]
	mov	cx, [cursor_x]
	mov	dx, [cursor_y]
	jmp	ctrl_proc_end
ctrl_proc_03h ENDP
;--------------------------------------- 04h: カーソル位置設定
ctrl_proc_04h PROC
	mov	si, bx			; save
	mov	ax, dx
	mov	bx, cx
	call	correct_cursor_pos
	cli
	mov	[cursor_x], bx
	mov	[cursor_y], ax
	mov	[cursor_mouse_x], bx
	mov	[cursor_mouse_y], ax
	mov	[cursor_tablet_x], bx
	mov	[cursor_tablet_y], ax
	mov	bx, si			; restore
	cmp	[cursor_counter], 0
	jnz	Lend
	call	draw_cursor_33h		; 描画中なら即座に描画を行う
Lend:	jmp	ctrl_proc_end
ctrl_proc_04h ENDP
;--------------------------------------- 05h: ボタン押下情報取得
ctrl_proc_05h PROC
	cmp	bx, 1
	ja	ctrl_proc_end
.if(zero?)
	mov	si, offset button_info_1
.else
	mov	si, offset button_info_0
.endif
	assume	si:ptr ButtonInfo
	xor	ah, ah
	cli
	mov	al, [button_status]
	mov	cx, [si].last_clk_x
	mov	dx, [si].last_clk_y
	mov	bx, [si].click_cnt
	mov	[si].click_cnt, 0
	assume	si:nothing
	jmp	ctrl_proc_end_ax_modify
ctrl_proc_05h ENDP
;--------------------------------------- 06h: ボタン解放情報取得
ctrl_proc_06h PROC
	cmp	bx, 1
	ja	ctrl_proc_end
.if(zero?)
	mov	si, offset button_info_1
.else
	mov	si, offset button_info_0
.endif
	assume	si:ptr ButtonInfo
	xor	ah, ah
	cli
	mov	al, [button_status]
	mov	cx, [si].last_rel_x
	mov	dx, [si].last_rel_y
	mov	bx, [si].release_cnt
	mov	[si].release_cnt, 0
	assume	si:nothing
	jmp	ctrl_proc_end_ax_modify
ctrl_proc_06h ENDP
;--------------------------------------- 07h: 水平方向範囲指定
ctrl_proc_07h PROC
	mov	ax, cx		; 後で復帰する ax, si で計算する
	mov	si, dx
.if(sword ptr ax < 0)
	xor	ax, ax
.elseif(sword ptr ax >= DISPLAY_WIDTH)
	mov	ax, DISPLAY_WIDTH - 1
.endif
.if(sword ptr si < 0)
	xor	si, si
.elseif(sword ptr si >= DISPLAY_WIDTH)
	mov	si, DISPLAY_WIDTH - 1
.endif
	cmp	ax, si
	jbe	@F
	xchg	ax, si
@@:	cli
	mov	[cursor_min_x], ax
	mov	[cursor_max_x], si
	jmp	ctrl_proc_end
ctrl_proc_07h ENDP
;--------------------------------------- 08h: 垂直方向範囲指定
ctrl_proc_08h PROC
	mov	ax, cx		; 後で復帰する ax, si で計算する
	mov	si, dx
.if(sword ptr ax < 0)
	xor	ax, ax
.elseif(sword ptr ax >= DISPLAY_HEIGHT)
	mov	ax, DISPLAY_HEIGHT - 1
.endif
.if(sword ptr si < 0)
	xor	si, si
.elseif(sword ptr si >= DISPLAY_HEIGHT)
	mov	si, DISPLAY_HEIGHT - 1
.endif
	cmp	ax, si
	jbe	@F
	xchg	ax, si
@@:	cli
	mov	[cursor_min_y], ax
	mov	[cursor_max_y], si
	jmp	ctrl_proc_end
ctrl_proc_08h ENDP
;--------------------------------------- 09h: カーソル形状指定
ctrl_proc_09h PROC
	cli
	; カーソル水平方向中心設定
.if(sword ptr bx < 0)
	mov	[cursor_center_x], 0
.elseif(sword ptr bx >= 16)
	mov	[cursor_center_x], 15
.else
	mov	[cursor_center_x], bx
.endif
	; カーソル垂直方向中心設定
.if(sword ptr cx < 0)
	mov	[cursor_center_y], 0
.elseif(sword ptr cx >= 16)
	mov	[cursor_center_y], 15
.else
	mov	[cursor_center_y], cx
.endif
	; カーソル形状コピー
	xchg	bx, dx				; save
	xor	si, si
L1:	mov	ax, es:[bx + si]
	xchg	ah, al
	mov	word ptr [user_cursor + si], ax
	add	si, 2
	cmp	si, 64
	jb	L1
	xchg	bx, dx				; restore
	or	[driver_flags], DRIVER_FLAG_USER_CURSOR
	cmp	[cursor_counter], 0
	jnz	Lend
	call	draw_cursor_33h			; 描画中なら即座に描画する
Lend:	jmp	ctrl_proc_end
ctrl_proc_09h ENDP
;--------------------------------------- 0bh: カーソル移動距離取得
ctrl_proc_0bh PROC
	cli
	mov	ax, [cursor_x]
	mov	si, [cursor_y]
	sti
	mov	cx, ax
	mov	dx, si
	sub	ax, [cursor_0bh_x]
	sub	si, [cursor_0bh_y]
	mov	[cursor_0bh_x], cx
	mov	[cursor_0bh_y], dx
	imul	[mky_dot_rate_x]
	SHFT_L	ax, 3
	mov	cx, ax
	mov	ax, si
	imul	[mky_dot_rate_y]
	SHFT_L	ax, 3
	mov	dx, ax
	jmp	ctrl_proc_end
ctrl_proc_0bh ENDP
;--------------------------------------- 0ch: ユーザ定義ルーチン呼び出し設定
ctrl_proc_0ch PROC
	mov	[user_subroutine + 0], dx
	mov	[user_subroutine + 2], es
	mov	[user_subroutine_flag], cl	; フラグ設定を最後にやる
	jmp	ctrl_proc_end
ctrl_proc_0ch ENDP
;--------------------------------------- 0fh: ミッキードット比設定
ctrl_proc_0fh PROC
	; 0 を設定しようとしていたら今の値をそのまま使う(idiv する為)
	test	cx, cx
	cli
	jz	@F
	mov	[mky_dot_rate_x], cx
@@:	test	dx, dx
	jz	@F
	mov	[mky_dot_rate_y], dx
@@:	sti
	jmp	ctrl_proc_end
ctrl_proc_0fh ENDP
;--------------------------------------- c0h: タブレット raw 値取得
ctrl_proc_c0h PROC
	cli
	mov	bx, [raw_pen_pressure]
	mov	cx, [raw_tablet_x]
	mov	dx, [raw_tablet_y]
	sti
	jmp	ctrl_proc_end
ctrl_proc_c0h ENDP
;--------------------------------------- c1h: マウス raw 値取得
ctrl_proc_c1h PROC
	xor	bh, bh
	xor	ch, ch
	xor	dh, dh
	cli
	mov	bl, [raw_mouse_btn]
	mov	cl, [raw_mouse_cnt_x]
	mov	dl, [raw_mouse_cnt_y]
	sti
	jmp	ctrl_proc_end
ctrl_proc_c1h ENDP
;--------------------------------------- c2h: カーソル描画カウンタ値取得
ctrl_proc_c2h PROC
	mov	dl, [cursor_counter]
	jmp	ctrl_proc_end
ctrl_proc_c2h ENDP

;---------------------------------------
; GDC の描画終了待機
; 引数, 戻値: 無し
wait_GDC PROC
@@:	in	al, 0a0h
	test	al, 08h
	jz	@F
	WAIT_5FH 2
	jmp	@B
@@:	ret
wait_GDC ENDP
;---------------------------------------
; IMR のマスク解除
; 引数: cl=マスク解除する IRQ
; 戻値: 無し
clear_imr_mask PROC
	push	dx
	mov	ah, 1
.if(cl <= 7)
	shl	ah, cl
	mov	dx, 02h
.else
	sub	cl, 8
	shl	ah, cl
	add	cl, 8
	mov	dx, 0ah
.endif
	not	ah
	cli
	in	al, dx
	and	al, ah
	out	dx, al
	sti
	pop	dx
	ret
clear_imr_mask ENDP
;---------------------------------------
; EOI 送信
; 引数, 戻値: 無し
; 破壊: al
send_EOI_to_slave PROC
	mov	al, 20h
	out	08h, al		; スレーブに送信
	WAIT_5FH 1
	mov	al, 0bh
	out	08h, al		; ISR 読み出しモード(スレーブ)
	WAIT_5FH 1
	in	al, 08h		; ISR 読む(スレーブ)
	test	al, al		; スレーブ PIC が in-service か判定
	jne	@F
send_EOI_to_master::
	mov	al, 20h
	out	00h, al		; マスタに送信
@@:	ret
send_EOI_to_slave ENDP

;---------------------------------------
; ボタン押下時の情報設定ルーチン
; 引数: CF=ボタン番号(0: 0, 1: 1), (bx,ax)=(x 位置, y 位置)
; 戻値: 無し
; 破壊: si
set_button_click_info PROC
	jc	Lbtn1
	mov	si, offset button_info_0
	jmp	@F
Lbtn1:	mov	si, offset button_info_1
@@:	assume	si:ptr ButtonInfo
	cli
	cmp	[si].click_cnt, 7fffh
	jae	@F
	inc	[si].click_cnt
@@:	mov	[si].last_clk_x, bx
	mov	[si].last_clk_y, ax
	sti
	assume	si:nothing
	ret
set_button_click_info ENDP
;-----------------------------
; ボタン解放時の情報設定ルーチン
; 引数: CF=ボタン番号(0: 0, 1: 1), (bx,ax)=(x 位置, y 位置)
; 戻値: 無し
; 破壊: si
set_button_release_info PROC
	jc	Lbtn1
	mov	si, offset button_info_0
	jmp	@F
Lbtn1:	mov	si, offset button_info_1
@@:	assume	si:ptr ButtonInfo
	cli
	cmp	[si].release_cnt, 7fffh
	jae	@F
	inc	[si].release_cnt
@@:	mov	[si].last_rel_x, bx
	mov	[si].last_rel_y, ax
	sti
	assume	si:nothing
	ret
set_button_release_info ENDP

;---------------------------------------
; カーソル消去ルーチン
; 引数, 戻値: 無し
; 破壊: ax, cx, di, si, es
erase_cursor PROC
	or	[driver_flags], DRIVER_FLAG_ERASING
	sti		; int 33h から cli した状態で呼ばれることがある為
	call	wait_GDC
	;--- カーソル消去処理本体
	mov	cx, 16
	mov	si, offset orig_vram
	mov	di, [last_draw_addr]
L1:	test	di, di
	js	next
	cmp	di, (DISPLAY_WIDTH / 8) * 400
	jae	next
	mov	ax, 0a800h
	call	restore_vram
	mov	ax, 0b000h
	call	restore_vram
	mov	ax, 0b800h
	call	restore_vram
	mov	ax, 0e000h
	call	restore_vram
next:	add	di, DISPLAY_WIDTH / 8
	loop	L1
	and	[driver_flags], NOT DRIVER_FLAG_ERASING
	ret
;--- vram 復元して si を進める
; 引数: ax=プレーンアドレス, di=書き込み先, si=復元対象データ
; 戻値: si=(si+3)
; 破壊: es
restore_vram:
	mov	es, ax
	mov	ax, [si]
	mov	es:[di], ax
	mov	al, [si + 2]
	mov	es:[di + 2], al
	add	si, 3
	ret
erase_cursor ENDP
;-----------------------------
; マスクデータ作成ルーチン
; 引数: bx=マスクデータオフセット, cl=シフト数
; 戻値: bp:bl=and mask シフト結果, si:bh=or mask シフト結果
; 破壊: ax, dx
make_mask_data PROC
	test	cl, cl
	jz	no_shift
	mov	dh, 8
	sub	dh, cl			; dh=cl-8
	;--- and mask
	mov	si, bx			; save
	mov	bx, 0ffffh		; マスク外ビットに1を立てる用
	mov	ax, [si]
	xchg	ah, al
	mov	dl, al
	shr	ax, cl
	shr	bl, cl
	xchg	cl, dh
	shl	dl, cl
	shl	bh, cl
	xchg	cl, dh
	or	dl, bl
	or	ah, bh
	xchg	ah, al
	mov	bp, ax
	mov	bl, dl
	;--- or mask
	mov	ax, [si + 32]
	xchg	ah, al
	mov	dl, al
	shr	ax, cl
	xchg	cl, dh
	shl	dl, cl
	xchg	cl, dh
	xchg	ah, al
	mov	si, ax
	mov	bh, dl
	ret
no_shift:
	mov	bp, [bx]		; and mask
	mov	si, [bx + 32]		; or mask
	mov	bx, 00ffh		; [ 15(or)8 | 7(and)0 ]
	ret
make_mask_data ENDP
;-----------------------------
; カーソル描画ルーチン(1ライン分)
; 引数: ax=プレーンアドレス, di=描画位置(バイト単位),
;       bp:bl=and mask, si:bh=or mask
; 戻値: 無し
; 破壊: ax, dl, es, ds(無条件に cs と等しくなる)
draw_cursor_l PROC
	mov	es, bx			; save bx
	mov	bx, [orig_vram_wptr]
	mov	ds, ax
	mov	ax, [di]
	mov	dl, [di + 2]
	mov	cs:[bx], ax		; 描画位置の vram 値を保存
	mov	cs:[bx + 2], dl
	add	bx, 3
	mov	cs:[orig_vram_wptr], bx
	mov	bx, es			; restore bx
	and	ax, bp
	and	dl, bl
	or	ax, si
	or	dl, bh
	mov	[di], ax
	mov	[di + 2], dl
	mov	ax, cs
	mov	ds, ax
	ret
draw_cursor_l ENDP
;-----------------------------
; カーソル描画ルーチン
; 引数, 戻値: 無し
; 破壊: 汎用レジスタ全部
draw_cursor PROC
	;--- 描画中フラグ操作
	cli
	mov	al, [driver_flags]
	test	al, (DRIVER_FLAG_DRAWING OR DRIVER_FLAG_ERASING)
	jz	L2
	test	al, DRIVER_FLAG_DRAWING
	jz	@F				; 描画中のみ再描画要求する
	or	[driver_flags], DRIVER_FLAG_REQ_REDRAW
@@:	sti
	ret
L2:	or	[driver_flags], DRIVER_FLAG_DRAWING
	sti
	push	es
	;--- vram 復元
	test	al, DRIVER_FLAG_FIRST_DRAW
	jnz	@F		; 初回描画なら復元不要
	call	erase_cursor	; GDC 描画完了待機は erase_cursor に任せる
	jmp	L1
@@:	call	wait_GDC	; erase_cursor しない場合ここで GDC 待機
L1:	;--- 描画準備
	and	[driver_flags], NOT DRIVER_FLAG_FIRST_DRAW
	mov	[orig_vram_wptr], offset orig_vram
	; 描画開始 X 位置計算
	mov	di, [cursor_x]
	mov	[cursor_last_draw_x], di
	mov	dx, [cursor_center_x]
	sub	di, dx
	mov	cx, di
	SHFT_R	di, 3		; di=バイト単位での x 方向の描画開始位置
	and	cl, 07h		; cl=左シフト量
	; 描画開始 Y 位置
	mov	ax, [cursor_y]
	mov	[cursor_last_draw_y], ax
	sub	ax, [cursor_center_y]	; ax=描画開始 Y 位置
	; Y 方向描画位置(X=0 とした場合)の VRAM 上でのアドレス
	mov	bx, DISPLAY_WIDTH / 8
	imul	bx		; dx:ax=ax*80
	mov	si, ax		; si=(X=0)での描画開始位置のアドレス
	;--- 描画ループ (ループ中保存: cl, ch, di, si, bx)
	add	ax, di
	mov	[last_draw_addr], ax
.if([driver_flags] & DRIVER_FLAG_USER_CURSOR)
	mov	bx, offset user_cursor
.else
	mov	bx, offset orig_cursor
.endif
	xor	ch, ch		; Y ループカウンタ
y_loop:
	;--- Y 方向描画範囲チェック
	test	si, si
	js	next_y
	cmp	si, (DISPLAY_WIDTH / 8) * DISPLAY_HEIGHT
	jae	next_y
	;--- X 方向描画範囲チェック
	push	di
	push	si
	push	bx
	mov	es, si			; save si
	cmp	di, -2
	jl	next_x
	je	set_mask_l2
	cmp	di, -1
	je	set_mask_l1
	cmp	di, (DISPLAY_WIDTH / 8) - 2
	je	set_mask_r2
	cmp	di, (DISPLAY_WIDTH / 8) - 1
	je	set_mask_r1
	jg	next_x
	;--- 1ライン分の描画処理
	; マスクデータ準備			; 描画開始位置=0～639-3の場合
	call	make_mask_data
	jmp	Lm_end
set_mask_l2:					; 描画開始位置=(0-2)の場合
	call	make_mask_data
	mov	bp, 0ffffh			; 冒頭2バイトを無効化
	xor	si, si
	jmp	Lm_end
set_mask_l1:					; 描画開始位置=(0-1)の場合
	call	make_mask_data
	or	bp, 000ffh			; 冒頭1バイトを無効化
	and	si, 0ff00h			; (little endian)
	jmp	Lm_end
set_mask_r2:					; 描画開始位置=(639-2)の場合
	call	make_mask_data
	mov	bl, 0ffh			; 末尾1バイトを無効化
	xor	bh, bh
	jmp	Lm_end
set_mask_r1:					; 描画開始位置=(639-1)の場合
	call	make_mask_data
	mov	bl, 0ffh			; 末尾2バイトを無効化
	xor	bh, bh
	or	bp, 0ff00h			; (little endian)
	and	si, 000ffh
Lm_end:; 描画位置計算
	mov	ax, es			; es=si
	add	di, ax
	; 描画
	mov	ax, 0a800h
	call	draw_cursor_l
	mov	ax, 0b000h
	call	draw_cursor_l
	mov	ax, 0b800h
	call	draw_cursor_l
	mov	ax, 0e000h
	call	draw_cursor_l
	;--- 1ライン分の描画処理ここまで
next_x:	pop	bx
	pop	si
	pop	di
next_y:	inc	ch
	cmp	ch, 16
	jae	@F
	add	bx, 2				; 次のマスクデータへ
	add	si, DISPLAY_WIDTH / 8		; 次のラインへ
	jmp	y_loop
@@:	;---
	pop	es
	and	[driver_flags], NOT DRIVER_FLAG_DRAWING
	ret
draw_cursor ENDP
;-----------------------------
; カーソル描画ルーチン(int 33h のハンドラから呼び出す用)
draw_cursor_33h PROC
	push	bx
	push	cx
	push	dx
	push	di
	push	bp
	call	draw_cursor
	pop	bp
	pop	di
	pop	dx
	pop	cx
	pop	bx
	ret
draw_cursor_33h ENDP
;---------------------------------------
; RS-232C の割込のハンドラ
; 注意: 処理中に追加でデータを受信するのを認めている為、serial_cnt や
;       serial_err に2命令以上使ってアクセスする場合は要割込禁止
;     - 処理中に serial_cnt が変化する可能性がある
;     - 処理中にエラーフラグが立つ可能性がある
serial_int_hander PROC
	sti
	push	ax
	push	dx
	push	cx
	push	bx
	push	si
	push	di
	push	bp
	push	ds
	mov	ax, cs
	mov	ds, ax
	;--- ステータス判定
	mov	bl, [com_line]
.if(bl >= COM_LINE_2ndCCU)
	xor	ah, ah
	mov	dx, [second_ccu_io_lsr]
	in	al, dx
	test	al, 04h				; PE
	jz	@F
	or	ah, SER_ERR_PE
@@:	test	al, 02h				; OE
	jz	@F
	or	ah, SER_ERR_OVE
@@:	test	al, 80h				; FE
	jz	@F
	or	ah, SER_ERR_FE
@@:	or	[serial_err], ah
	; Data Ready 以外なら何もしない
	test	al, 01h				; Data Ready
	jz	end_eoi
.else
	mov	dx, [serial_io_cmd]
	in	al, dx
	mov	ah, al
	and	al, 38h			; エラー部分を取り出す
	or	[serial_err], al
	; RxRDY 以外なら何もしない
	test	ah, 02h
	jz	end_eoi
.endif
	;--- 既に読んだデータ数チェック
	cmp	[serial_cnt], SERIAL_BUF_LEN
	jb	@F
	or	[serial_err], SER_ERR_TSR_OVE
	jmp	end_eoi
@@:	;--- データを読み込んでバッファに入れる(bl=[com_line])
	inc	[serial_total]
.if(bl >= COM_LINE_2ndCCU)
	mov	dx, [second_ccu_io_data]
.else
	mov	dx, [serial_io_dat]
.endif
	in	al, dx
	cmp	[serial_mode], SER_MODE_NO_OPERATE
	je	end_eoi					; 何もしない場合
	mov	bx, [serial_wptr]
	mov	[bx], al
	inc	bx
	cmp	bx, offset serial_buf + SERIAL_BUF_LEN
	jb	@F
	mov	bx, offset serial_buf
@@:	mov	[serial_wptr], bx
	inc	[serial_cnt]
	;--- 処理中フラグ確認
	test	[serial_flags], SER_FLAG_PROCESSING
	jnz	end_eoi
	or	[serial_flags], SER_FLAG_PROCESSING
	; EOI 送って割込許可
	call	[serial_eoi_func]
	sti			; これ以降割込を許可(データ取りこぼし防止)
	;--- モードに合わせた処理をする
	xor	bh, bh
	mov	bl, [serial_mode]
	add	bx, bx			; word access
	jmp	[tbl_serial_proc + bx]
serial_int_handler_end::
	and	[serial_flags], NOT SER_FLAG_PROCESSING
	;---
end_3:	pop	ds
	pop	bp
	pop	di
	pop	si
	pop	bx
	pop	cx
	pop	dx
	pop	ax
	iret
	;--- EOI を送って終了する
end_eoi:
	; パース中にエラーがあればバッファを全部破棄してエラークリア
	cmp	[serial_mode], SER_MODE_PARSE_DATA
	jnz	@F
	mov	ax, [serial_wptr]
	mov	[serial_rptr], ax
	mov	[serial_cnt], 0
	mov	[serial_err], SER_NO_ERR
@@:	call	[serial_eoi_func]
	jmp	end_3
serial_int_hander ENDP

;--- '~' から 0dh までのデータを保存する
; 引数: ds:dx=これまで読み取ったバイト数(word)
;       ds:bx=保存先バッファ
;       cx=保存先バッファのサイズ
; 戻値: 無し
; 破壊: ax, si, di
save_tilde_to_cr PROC
	mov	si, dx
	mov	di, [si]
	mov	si, [serial_rptr]
L1:	cmp	[serial_cnt], 0
	jnz	@F
	mov	[serial_rptr], si
	mov	si, dx
	mov	[si], di
	ret
@@:	mov	al, [si]
	dec	[serial_cnt]
	inc	si
	cmp	si, offset serial_buf + SERIAL_BUF_LEN
	jb	@F
	mov	si, offset serial_buf
@@:	; 保存処理本体
	mov	ah, [serial_flags]
	test	ah, SER_FLAG_READ_OK
	jnz	L1			; 既に OK なら読み直しはしない
	test	ah, SER_FLAG_READING
	jz	not_reading
	cmp	al, 0dh
	jnz	L2			; 末尾の 0dh は保存しない
	and	ah, NOT SER_FLAG_READING
	or	ah, SER_FLAG_READ_OK
	mov	[serial_flags], ah
	mov	[serial_mode], SER_MODE_NO_OPERATE
	jmp	L1
not_reading:
	cmp	al, '~'
	jnz	L1
	or	[serial_flags], SER_FLAG_READING
L2:	cmp	di, cx
	jb	@F
	or	[serial_err], SER_ERR_BUF_FULL
	jmp	L1
@@:	mov	[bx + di], al
	inc	di
	jmp	L1
save_tilde_to_cr ENDP
ser_proc_get_mdoel PROC
	mov	bx, offset tablet_name
	mov	dx, offset tablet_name_len
	mov	cx, TABLET_NAME_BUF_LEN
	mov	ax, serial_int_handler_end ; save_tilde_to_cr の ret で戻る
	push	ax
	jmp	save_tilde_to_cr
ser_proc_get_mdoel ENDP
ser_proc_get_area PROC
	mov	bx, offset tablet_area_str
	mov	dx, offset tablet_area_str_len
	mov	cx, TABLET_AREA_STR_BUF_LEN
	mov	ax, serial_int_handler_end
	push	ax
	jmp	save_tilde_to_cr
ser_proc_get_area ENDP
;---
ser_proc_parse_data PROC
	;--- ループ中 si は保持する必要あり
	mov	si, [serial_rptr]
Ltop:	cmp	[serial_cnt], 0
	jnz	@F
	mov	[serial_rptr], si
	jmp	serial_int_handler_end
@@:	mov	al, [si]
	dec	[serial_cnt]
	inc	si
	cmp	si, offset serial_buf + SERIAL_BUF_LEN
	jb	@F
	mov	si, offset serial_buf
@@:	;--- パース用バッファにデータ移動
	test	al, 80h
	jz	@F
	mov	[tablet_packet_buf], al
	mov	[tablet_packet_n_recv], 1
	jmp	Ltop
@@:	xor	bh, bh
	mov	bl, [tablet_packet_n_recv]
	mov	[tablet_packet_buf + bx], al
	inc	bl
	cmp	bl, [tablet_packet_len]
	jae	@F
	mov	[tablet_packet_n_recv], bl
	jmp	Ltop
@@:	;--- パケットパース
	mov	[tablet_packet_n_recv], 0
.if([tablet_proto] == PROTO_W_IV)
	;--- protocol iv
	; MSB=1 と接近の判定
	mov	ax, word ptr [tablet_packet_buf]
	mov	dl, al
	and	dl, 0c0h
	cmp	dl, 0c0h
	jnz	Ltop
	; X 座標
	xchg	al, ah
	shl	al, 1
	SHFT_L	ax, 6
	mov	dx, word ptr [tablet_packet_buf + 2]
	or	al, dl
	mov	[raw_tablet_x], ax
	; ボタン押下状態(ボタン0は筆圧で取るためボタン1のみ判定)
	mov	al, dh
	and	al, 10h		; ボタン0もここで判定する場合は 18h を使う
	SHFT_R	al, 3
	mov	[raw_tablet_btn], al
	; Y 座標
	xor	dl, dl
	mov	ax, word ptr [tablet_packet_buf + 4]
	xchg	al, ah
	shl	al, 1
	shr	ax, 1
	SHFT_L	dh, 6
	rcl	dl, 1			; bit3 を dl に回収
	or	ah, dh
	mov	[raw_tablet_y], ax
	; 筆圧
	mov	al, [tablet_packet_buf + 6]
	shl	al, 1
	or	al, dl			; bit3 をつける
	cmp	al, -120
	jge	@F
	mov	al, -120
	jmp	L2
@@:	cmp	al, 120
	jle	L2
	mov	al, 120
L2:	add	al, 120			; 0～240の範囲に変換
	xor	ah, ah
	mov	[raw_pen_pressure], ax
.else
	;--- protocol v
	mov	ax, word ptr [tablet_packet_buf]
	mov	dx, ax				; save
	and	al, 0b8h			; 一応 MSB=1 も確認
	cmp	al, 0a0h
	jnz	Ltop			; この場合ペン以外の模様
	; ボタン押下状態
	xor	dh, dh
	test	dl, 06h			; どちらのボタンも右クリック扱い
	jz	@F
	or	dh, MOUSE_BUTTON_1
@@:	mov	[raw_tablet_btn], dh
	; X 座標
	xor	dx, dx
	mov	dh, ah
	mov	ax, word ptr [tablet_packet_buf + 2]
	xchg	ah, al
	shl	al, 1
	mov	ch, al					; save
	SHFT_L	ax, 2
	rcl	dh, 1
	mov	dl, ah
	mov	[raw_tablet_x], dx
	; Y 座標
	SHFT_L	ch, 2
	mov	ax, word ptr [tablet_packet_buf + 4]
	mov	dh, ah					; save
	xchg	al, ah
	shl	al, 1
	SHFT_R	ax, 4
	or	ah, ch
	mov	[raw_tablet_y], ax
	; 筆圧
	and	dh, 07h
	mov	dl, [tablet_packet_buf + 6]
	shl	dl, 1
	shr	dx, 1
	mov	[raw_pen_pressure], dx
.endif
	;--- タブレットから取得した生の値の変換
	; 筆圧によるクリック判定
	mov	dl, [raw_tablet_btn]
	mov	[raw_tablet_btn_old], dl
	mov	ax, [raw_pen_pressure]
	cmp	ax, [tablet_p_threshold]
	jb	@F
	or	dl, MOUSE_BUTTON_0
	mov	[raw_tablet_btn], dl
@@:	; 取得した X, Y 座標をマウス座標に変換
	; X 座標変換
	mov	ax, [raw_tablet_x]
	mov	dx, [tablet_margin_l]
	mov	cx, [tablet_real_width]
	cmp	ax, dx
	jae	@F
	xor	ax, ax
	jmp	L3
@@:	sub	ax, dx
	cmp	ax, cx
	jb	L3
	mov	ax, cx
L3:	mov	dx, DISPLAY_WIDTH
	mul	dx
	div	cx
	mov	bx, ax			; save
	; Y 座標変換
	mov	ax, [raw_tablet_y]
	mov	dx, [tablet_margin_t]
	mov	cx, [tablet_real_height]
	cmp	ax, dx
	jae	@F
	xor	ax, ax
	jmp	L4
@@:	sub	ax, dx
	cmp	ax, cx
	jb	L4
	mov	ax, cx
L4:	mov	dx, DISPLAY_HEIGHT
	mul	dx
	div	cx
	; 補正して保存
	call	correct_cursor_pos
	mov	[cursor_tablet_x], bx
	mov	[cursor_tablet_y], ax
	; 全部設定し終わったら最後にフラグを立てる
	or	[driver_flags], DRIVER_FLAG_UPD_TABLET
	jmp	Ltop
ser_proc_parse_data ENDP

;-----------------------------
; マウスカーソルの位置を指定された移動範囲内に収まるように補正する
; 引数: (bx,ax)=(X 座標, Y座標)
; 戻値: (bx,ax)=(補正後 X 座標, 補正後 Y座標)
correct_cursor_pos PROC
.if(sword ptr bx < [cursor_min_x])
	mov	bx, [cursor_min_x]
.elseif(sword ptr bx > [cursor_max_x])
	mov	bx, [cursor_max_x]
.endif
.if(sword ptr ax < [cursor_min_y])
	mov	ax, [cursor_min_y]
.elseif(sword ptr ax > [cursor_max_y])
	mov	ax, [cursor_max_y]
.endif
	ret
correct_cursor_pos ENDP

;---------------------------------------
; マウスのハードウェア割込関連処理
;-----------------------------
; マウス 8255A ポートA から値を読むマクロ
; out_val: ポート C への出力値, reg: 読んだ値を保存するレジスタ
IN_MOUSE_PORT_A MACRO out_val:REQ, reg:REQ
	mov	dx, MOUSE_IO_PORT_C
	mov	al, out_val
	out	dx, al
	WAIT_5FH 2
	mov	dx, MOUSE_IO_PORT_A
	in	al, dx
IFDIFI <reg>, <al>			; reg が al なら mov 省略
	mov	reg, al
ENDIF
	WAIT_5FH 1
ENDM
;-----------------------------
; マウスカウント値ドット値変換
; 引数: al=マウスカウント値, dx=前回割込時のミッキー値の余り, si=mky/dot rate
; 戻値: ax=変換後のドット値, dx=ミッキー値の余り
; #DE 例外を出さないため被除数の範囲は 8000h～7fffh (の符号拡張)に制限する
conv_mouse_cnt_to_dot PROC
	cbw			; ax=(符号拡張)al
	SHFT_L	ax, 3
	add	ax, dx		; ax=前回割込からの総ミッキー値
	jno	L1		; 桁溢れ対応(多分溢れないと思うが)
	jns	@F
	mov	ax, 7fffh
	jmp	L1
@@:	mov	ax, 8000h
L1:	cwd			; dx:ax=符号拡張(ax)
	idiv	si		; dx:ax/rate -> ax(商), dx(余)
	ret
conv_mouse_cnt_to_dot ENDP
;-----------------------------
; マウスのハードウェア割込のハンドラ
mouse_int_hander PROC
	;--- 処理中フラグ操作
	test	cs:[mouse_flags], MOUSE_FLAG_PROCESSING
	jz	@F
	push	ax
	call	send_EOI_to_slave
	pop	ax
	iret
@@:	or	cs:[mouse_flags], MOUSE_FLAG_PROCESSING
	push	ds
	push	bp
	push	di
	push	si
	push	bx
	push	cx
	push	dx
	push	ax
	;--- EOI 送信
	call	send_EOI_to_slave
	sti
	;--- データ読み取り
	mov	ax, cs
	mov	ds, ax
	IN_MOUSE_PORT_A 0a0h, bh	; X 上位 (HC=1, SXY=0, SHL=1)
	IN_MOUSE_PORT_A 080h, bl	; X 下位 (HC=1, SXY=0, SHL=0)
	IN_MOUSE_PORT_A 0e0h, ah	; Y 上位 (HC=1, SXY=1, SHL=1)
	IN_MOUSE_PORT_A 0c0h, al	; Y 下位 (HC=1, SXY=1, SHL=1)
	;--- データ整形
	mov	dl, al			; save
	SHFT_L	ah, 4
	and	al, 0fh
	or	ah, al			; ah=Y 方向カウンタ値
	mov	al, bl
	SHFT_L	bh, 4
	and	al, 0fh
	or	al, bh			; al=X 方向カウンタ値
	xor	dh, dh
	test	dl, 80h
	jnz	@F
	or	dh, MOUSE_BUTTON_0
@@:	test	dl, 20h
	jnz	@F
	or	dh, MOUSE_BUTTON_1
@@:	;--- 保存
	mov	dl, [raw_mouse_btn]
	mov	[raw_mouse_btn_old], dl
	mov	[raw_mouse_btn], dh
	mov	word ptr [raw_mouse_cnt_x], ax
	test	ax, ax	; 動いているか判定
	jnz	@F
.if(dl != dh)		; 全く動いていない場合ボタン押下の変化をチェック
	or	[driver_flags], DRIVER_FLAG_UPD_MOUSE
.endif
	mov	bx, [cursor_mouse_x]
	mov	ax, [cursor_mouse_y]
	xor	si, si
	mov	[remain_mky_x], si	; ミッキー値の余りをクリア
	mov	[remain_mky_y], si
	jmp	no_move
@@:	;--- マウスカウント値ドット値変換
	or	[driver_flags], DRIVER_FLAG_UPD_MOUSE
	mov	bx, ax			; save
	; X 方向
	mov	dx, [remain_mky_x]
	mov	ah, al
	xor	ah, dh
	jns	@F
	xor	dx, dx	; マウスカウント値と余りの符号が逆なら余り=0にする
@@:	mov	si, [mky_dot_rate_x]
	call	conv_mouse_cnt_to_dot
	mov	[remain_mky_x], dx
	add	ax, [cursor_mouse_x]	; ax=新カーソル位置(X 方向)
	; Y 方向
	xchg	bx, ax			; bx=結果(X 方向), ah=raw_mouse_cnt_y
	mov	dx, [remain_mky_y]
	mov	al, ah
	xor	ah, dh
	jns	@F
	xor	dx, dx
@@:	mov	si, [mky_dot_rate_y]
	call	conv_mouse_cnt_to_dot
	mov	[remain_mky_y], dx
	add	ax, [cursor_mouse_y]	; ax=新カーソル位置(Y 方向)
no_move:
	call	correct_cursor_pos
	; 座標計算結果設定 ((bx,ax)=(x 位置,y 位置))
	mov	[cursor_mouse_x], bx
	mov	[cursor_mouse_y], ax
	;--- ボタン押下状態計算
	xor	al, al
	mov	dl, [raw_mouse_btn]
	mov	dh, [raw_tablet_btn]
	test	dx, (MOUSE_BUTTON_0 SHL 8) OR MOUSE_BUTTON_0
	jz	@F
	or	al, MOUSE_BUTTON_0
@@:	test	dx, (MOUSE_BUTTON_1 SHL 8) OR MOUSE_BUTTON_1
	jz	@F
	or	al, MOUSE_BUTTON_1
@@:	mov	ah, [button_status]
	mov	word ptr [button_status], ax	; 保存
	; ユーザ定義ルーチン呼び出し判定(ボタン)
	xor	dl, dl		; ユーザ定義サブルーチン呼び出しフラグ
	xor	ah, al		; 押下状態変化していたら1
.if(ah & MOUSE_BUTTON_0)
  .if(al & MOUSE_BUTTON_0)
	or	dl, USER_SUBROUTINE_L_CLICKED
  .else
	or	dl, USER_SUBROUTINE_L_RELEASED
  .endif
.endif
.if(ah & MOUSE_BUTTON_1)
  .if(al & MOUSE_BUTTON_1)
	or	dl, USER_SUBROUTINE_R_CLICKED
  .else
	or	dl, USER_SUBROUTINE_R_RELEASED
  .endif
.endif
	;--- カーソル位置計算
	mov	cl, [driver_flags]
	test	cl, (DRIVER_FLAG_UPD_TABLET OR DRIVER_FLAG_UPD_MOUSE)
	jz	no_cursor_upd
	; 採用するカーソル位置はタブレット優先
.if(cl & DRIVER_FLAG_UPD_TABLET)
	mov	bx, [cursor_tablet_x]
	mov	ax, [cursor_tablet_y]
	and	[driver_flags], NOT DRIVER_FLAG_UPD_TABLET
  .if([driver_flags_s] & DRIVER_FLAG_S_SYNC_CURSOR)
	mov	[cursor_mouse_x], bx
	mov	[cursor_mouse_y], ax
  .endif
.else
	mov	bx, [cursor_mouse_x]
	mov	ax, [cursor_mouse_y]
	and	[driver_flags], NOT DRIVER_FLAG_UPD_MOUSE
  .if([driver_flags_s] & DRIVER_FLAG_S_SYNC_CURSOR)
	cli
	mov	[cursor_tablet_x], bx
	mov	[cursor_tablet_y], ax
	sti
  .endif
.endif
.if((bx != [cursor_x]) || (ax != [cursor_y]))
	or	dl, USER_SUBROUTINE_POS_CHANGE
.endif
	mov	[cursor_x], bx
	mov	[cursor_y], ax
no_cursor_upd:
	;--- ボタン押下/解放情報更新
	test	dl, USER_SUBROUTINE_L_CLICKED
	jz	@F
	clc
	call	set_button_click_info
@@:	test	dl, USER_SUBROUTINE_L_RELEASED
	jz	@F
	clc
	call	set_button_release_info
@@:	test	dl, USER_SUBROUTINE_R_CLICKED
	jz	@F
	stc
	call	set_button_click_info
@@:	test	dl, USER_SUBROUTINE_R_RELEASED
	jz	@F
	stc
	call	set_button_release_info
@@:	;--- カーソル描画
.if([cursor_counter] == 0)
	; 描画条件: ドット単位で動きがある、初回描画、スキップされている
	test	dl, USER_SUBROUTINE_POS_CHANGE
	jnz	@F
	test	[driver_flags], \
			(DRIVER_FLAG_FIRST_DRAW OR DRIVER_FLAG_REQ_REDRAW)
	jz	L1
@@:	and	[driver_flags], NOT DRIVER_FLAG_REQ_REDRAW
	push	dx
	call	draw_cursor
	pop	dx
.endif
L1:	;--- ユーザ定義サブルーチン呼び出し
	test	dl, [user_subroutine_flag]
	jz	skip_user_subr
	xor	dh, dh
	mov	di, dx				; save
	mov	ax, [cursor_x]
	mov	cx, ax
	sub	ax, [cursor_user_subr_x]
	mov	[cursor_user_subr_x], cx
	imul	[mky_dot_rate_x]
	SHFT_L	ax, 3
	mov	si, ax
	mov	ax, [cursor_y]
	mov	bx, ax				; save
	sub	ax, [cursor_user_subr_y]
	mov	[cursor_user_subr_y], bx
	imul	[mky_dot_rate_y]
	SHFT_L	ax, 3
	xchg	di, ax				; ax=呼び出しフラグ
	mov	dx, bx				; restore
	xor	bh, bh
	mov	bl, [button_status]
	cli
	call	dword ptr [user_subroutine]
	sti
skip_user_subr:
	;--- 後始末
	mov	dx, MOUSE_IO_PORT_C
	xor	al, al			; HC=0 に戻す
	out	dx, al
	;---
	pop	ax
	pop	dx
	pop	cx
	pop	bx
	pop	si
	pop	di
	pop	bp
	and	[mouse_flags], NOT MOUSE_FLAG_PROCESSING
	pop	ds
	iret
mouse_int_hander ENDP

;---------------------------------------
; 常駐するデータ
;---------------------------------------
	even
tsr_data_head:
orig_vec_ctrl	dw ?, ?		; off, seg
orig_vec_mouse	dw ?, ?
orig_vec_serial	dw ?, ?

serial_eoi_func	dw offset send_EOI_to_master	; RS-232C 割込の EOI 送信関数

tbl_ctrl_proc dw	offset ctrl_proc_00h, offset ctrl_proc_01h,
			offset ctrl_proc_02h, offset ctrl_proc_03h,
			offset ctrl_proc_04h, offset ctrl_proc_05h,
			offset ctrl_proc_06h, offset ctrl_proc_07h,
			offset ctrl_proc_08h, offset ctrl_proc_09h,
			offset ctrl_proc_dummy, offset ctrl_proc_0bh,
			offset ctrl_proc_0ch, offset ctrl_proc_dummy,
			offset ctrl_proc_dummy, offset ctrl_proc_0fh
tbl_ctrl_ext_proc dw	offset ctrl_proc_c0h, offset ctrl_proc_c1h,
			offset ctrl_proc_c2h
tbl_serial_proc dw	0, ; dummy for SER_MODE_NO_OPERATE
			offset ser_proc_get_mdoel,
			offset ser_proc_get_area,
			offset ser_proc_parse_data

last_draw_addr	dw ?			; 前回カーソルを描画した vram アドレス
orig_vram_wptr	dw ?
orig_vram 	db (3*16*4) dup(?)	; カーソル描画前の元々の vram の値
; 内蔵マウスカーソルデータ
orig_cursor db	01111111b, 11111111b,	; and mask
		00111111b, 11111111b,
		00011111b, 11111111b,
		00001111b, 11111111b,
		00000111b, 11111111b,
		00000011b, 11111111b,
		00000001b, 11111111b,
		00000000b, 11111111b,
		00000000b, 01111111b,
		00000001b, 11111111b,
		00000011b, 11111111b,
		00110001b, 11111111b,
		01110001b, 11111111b,
		11111000b, 11111111b,
		11111000b, 11111111b,
		11111100b, 11111111b

	db	00000000b, 00000000b,	; xor mask
		00000000b, 00000000b,
		01000000b, 00000000b,
		01100000b, 00000000b,
		01110000b, 00000000b,
		01111000b, 00000000b,
		01111100b, 00000000b,
		01111110b, 00000000b,
		01111110b, 00000000b,
		01111000b, 00000000b,
		01001000b, 00000000b,
		00000100b, 00000000b,
		00000100b, 00000000b,
		00000010b, 00000000b,
		00000010b, 00000000b,
		00000000b, 00000000b
user_cursor db 64 dup(?)

cursor_tablet_x		dw ?	; raw からの変換値
cursor_tablet_y		dw ?
cursor_mouse_x		dw ?
cursor_mouse_y		dw ?
cursor_0bh_x		dw ?	; 前回の ax=0bh での座標の記録用
cursor_0bh_y		dw ?
cursor_user_subr_x	dw ?	; 前回のユーザー定義サブルーチン呼び出し時の
cursor_user_subr_y	dw ?	; 座標の記録用
cursor_last_draw_x	dw ?	; 前回カーソル描画を行った位置
cursor_last_draw_y	dw ?

init_zero_head:					; == 初期化時0クリア範囲先頭 ==
; タブレットのパケットから取得した生の値
raw_tablet_x		dw ?
raw_tablet_y		dw ?
raw_pen_pressure	dw ?
raw_tablet_btn		db ?	; 注: word access する場合有り
raw_tablet_btn_old	db ?
; マウスから取得した生の値
raw_mouse_btn		db ?	; 注: word access する場合有り
raw_mouse_btn_old	db ?
raw_mouse_cnt_x		db ?	; 注: word access する場合有り
raw_mouse_cnt_y		db ?
; カーソル関係 (注意: 最大値は (DISPLAY_WIDTH - 1), (DISPLAY_HEIGHT - 1))
cursor_center_x		dw 0	; カーソル中心点座標
cursor_center_y		dw 0
cursor_x		dw ?	; 制御割込で返す座標値
cursor_y		dw ?
cursor_min_x		dw ?
cursor_max_x		dw ?
cursor_min_y		dw ?
cursor_max_y		dw ?
cursor_counter		db ?

; ユーザ定義サブルーチン
user_subroutine_flag	db 0		; 呼び出し条件
user_subroutine		dw ?, ?		; far address

; ボタン関係
button_status		db ?	; ボタン押下状態
button_status_old	db ?
button_info_0	ButtonInfo { }
button_info_1	ButtonInfo { }

; ミッキー値関係
remain_mky_x		dw ?	; 前回のミッキー/ドット変換での余り
remain_mky_y		dw ?
init_zero_tail:					; == 初期化時0クリア範囲末尾 ==
mky_dot_rate_x		dw MKY_DOT_RATE_INIT
mky_dot_rate_y		dw MKY_DOT_RATE_INIT
mky_dot_rate_specify	dw MKY_DOT_RATE_INIT	; /k で指定されたミッキー値

; 0クリア範囲のサイズチェックマクロ
IF (offset init_zero_tail - offset init_zero_head) MOD 2 EQ 1
	.err <Size of the init_zero area is not odd.>
ENDIF

tablet_model		dw ?
tablet_p_threshold	dw ?	; クリック判定の発生する筆圧の閾値
tablet_area_width	dw ?	; タブレット座標範囲
tablet_area_height	dw ?
tablet_real_width	dw ?	; 余白幅を考慮した実際の座標範囲
tablet_real_height	dw ?
; タブレット感知域の上下左右の余白幅
; /m のパースで連続アクセスするので tblr の順を変えてはならない
tablet_margin_t		dw TABLET_DEFAULT_MARGIN
tablet_margin_b		dw TABLET_DEFAULT_MARGIN
tablet_margin_l		dw TABLET_DEFAULT_MARGIN
tablet_margin_r		dw TABLET_DEFAULT_MARGIN

; 8251A 制御 IO アドレス
serial_io_dat	dw 30h		; data register
serial_io_cmd	dw 32h		; mode set / command word write
serial_io_sys	dw 35h		; system port

; 16550 制御 IO アドレス
; 順番を入れ替える場合 set_16550_io_ports も変更すること
second_ccu_io_data	dw 238h	; データレジスタ
second_ccu_io_dl	dw 238h	; 分周比レジスタ(下位)
second_ccu_io_ier	dw 239h	; 割込許可レジスタ
second_ccu_io_dh	dw 239h	; 分周比レジスタ(上位)
second_ccu_io_iir	dw 23ah	; 割込参照レジスタ
second_ccu_io_fcr	dw 23ah	; FIFO 制御レジスタ
second_ccu_io_lcr	dw 23bh	; ラインコントロールレジスタ
second_ccu_io_mcr	dw 23ch	; モデムコントロールレジスタ
second_ccu_io_lsr	dw 23dh	; ラインステータスレジスタ
second_ccu_io_msr	dw 23eh	; モデムステータスレジスタ
second_ccu_io_scr	dw 23fh	; スクラッチレジスタ


; RS-232C リングバッファ
SERIAL_BUF_LEN	equ 64
serial_cnt	dw 0				; バッファ中のデータ数
serial_total	dw 0				; 累計読み出し数
serial_wptr	dw offset serial_buf		; serial_buf 書き込み位置
serial_rptr	dw offset serial_buf		; serial_buf 書き込み位置
serial_buf	db SERIAL_BUF_LEN dup(?)	; RS-232C 受信データバッファ
; タブレットモデル名バッファ
TABLET_NAME_BUF_LEN	equ 32
tablet_name_len		dw 0
tablet_name		db TABLET_NAME_BUF_LEN dup(0)	; NULL 終端の為
; タブレット座標範囲バッファ
TABLET_AREA_STR_BUF_LEN	equ 20
tablet_area_str_len	dw 0
tablet_area_str		db TABLET_AREA_STR_BUF_LEN dup(0)
; プロトコル
tablet_proto		db ?
tablet_packet_len	db ?
tablet_packet_buf	db LONGEST_PACKET_LEN dup(?)
tablet_packet_n_recv	db 0

serial_mode	db SER_MODE_NO_OPERATE
serial_err	db SER_NO_ERR
serial_flags	db 0			; RS-232 処理関係フラグ

driver_flags	db 0				; ドライバ制御フラグ
driver_flags_s	db DRIVER_FLAG_S_INIT_VAL	; やや static なフラグ

mouse_int_freq	db 0		; マウス割込頻度(MOUSE_IO_INT_FREQ に書く値)
mouse_flags	db 0		; マウス処理関係フラグ

int_mouse	db 6		; INT6
irq_mouse	db ?		; int_mose から計算
int_serial	db 0ffh		; 0ffh=IRQ4 (標準第1回線)
irq_serial	db 4		; int_serial != 0ffh なら計算

IMR_FLAG_SERIAL	equ 01h
IMR_FLAG_MOUSE	equ 02h
orig_imr_mask	db 0		; IMR がもともとマスクされていたかどうか

com_line	db COM_LINE_1st	; 通信回線

tsr_tail:

;-----------------------------------------------------------
; 非常駐部分
;---------------------------------------
	even
tbl_err_msg	dw 0, ; dummy for PROG_NO_ERR
		offset msg_perr_arg,
		offset msg_perr_tablet_setup,
		offset msg_perr_already_stayed,
		offset msg_perr_not_stayed,
		offset msg_perr_invalid_margin,
		offset msg_perr_no_2ndccu,
		offset msg_perr_no_9861k,
		offset msg_perr_no_rsa_98ii

tbl_mode_proc	dw program_end, ; dummy for PROG_MODE_NO_OPERATE
		do_stay_j, do_remove_j, do_help_j

err_add_msg	dw 0		; エラー時に表示する追加メッセージ
io_offset_9861k	dw 0		; 9861K の IO アドレスのオフセット
specified_addr	dw 238h		; /a で指定された IO アドレス
prog_err	db PROG_NO_ERR	; プログラム実行時のエラー
prog_flags	db 0		; プログラム実行に関係するフラグ
prog_mode	db PROG_MODE_NO_OPERATE	; プログラムの動作モード
specified_opts	db 0		; プログラム引数で指定済みのオプション
specified_int	db 1		; /i で指定された INT

INCLUDE msg.inc

ser_data_init_and_get_model db	0dh, 0dh, 0dh, '#', 0dh,
				0dh, 0dh, 0dh, '~#', 0dh
ser_data_get_tablet_area_xd db 0dh, 0dh, 0dh, 'ST', 0dh
ser_data_get_tablet_area db 0dh, 0dh, 0dh, '~C', 0dh

;---------------------------------------
; エントリポイント
main PROC
	mov	sp, stack_start
	cld
	mov	ax, offset msg_header
	call	putstr
	;--- プログラム引数パース
	call	parse_args
	jc	program_end
	xor	bh, bh
	mov	bl, [prog_mode]
	add	bx, bx		; word access
	jmp	word ptr [tbl_mode_proc + bx]
	;--- プログラム終了処理
program_end::
	mov	bl, [prog_err]
.if(bl != PROG_NO_ERR)
	; エラーメッセージ表示
	mov	ax, offset msg_header_err
	call	putstr
	xor	bh, bh
	add	bx, bx		; word access
	mov	ax, [tbl_err_msg + bx]
	call	putstr_newline
	mov	ax, [err_add_msg]
	test	ax, ax
	jz	@F
	call	putstr_newline
@@:	mov	al, 1		; 終了コード
.else
	xor	al, al
.endif
	test	[prog_flags], PROG_FLAG_DO_STAY
	jnz	@F
	mov	ah, 4ch
	int	21h
@@:	mov	ah, 31h
	mov	dx, offset tsr_tail
	SHFT_R	dx, 4
	inc	dx
	int	21h
main ENDP


;---------------------------------------
; タブレットセットアップ失敗扱いにするマクロ
FAILED_TO_SETUP_TABLET MACRO msg:REQ
	mov	[err_add_msg], offset msg
	jmp	failed_on_tablet
ENDM
;--- 常駐処理
do_stay_j PROC
	;--- プログラム引数のパースが完了しないとできないオプションの確認
	mov	al, [com_line]
.if(al == COM_LINE_9861K_2nd || al == COM_LINE_9861K_3rd)
	; 指定可能アドレスは 00h～0hf の範囲
	test	[specified_opts], OPT_IO_ADDR
	jnz	@F
	mov	[specified_addr], 0	; デフォルト値ロード
@@:	cmp	[specified_addr], 0fh
	jbe	@F
	mov	[prog_err], PROG_ERR_ARG
	jmp	program_end
.elseif(al == COM_LINE_RSA_98II)
	; RSA-98III も考慮して下位 3bit が0かのみチェック
	test	[specified_addr], 07h
	jz	@F
	mov	[prog_err], PROG_ERR_ARG
	jmp	program_end
.endif
@@:	;--- 常駐チェック
	call	check_stay
	jnc	@F
	mov	[prog_err], PROG_ERR_ALREADY_STAYED
	jmp	program_end
@@:	;--- 引数で指定された回線から INT 取得
	mov	al, [com_line]
.if(al == COM_LINE_2ndCCU)
	; 2ndCCU が存在するかチェック(F8E8:0011h bit3)
	mov	ax, 0f8e8h
	mov	es, ax
	xor	dl, dl
	test	byte ptr es:[11h], 08h
	jz	@F
	inc	dl
@@:	mov	ax, cs
	mov	es, ax
	test	dl, dl
	jnz	@F
	mov	[prog_err], PROG_ERR_NO_2NDCCU
	jmp	program_end
@@:	mov	[int_serial], 1		; INT 1
.elseif(al == COM_LINE_RSA_98II)
	; 指定したアドレスにデバイスがあるかチェック
	mov	dx, [specified_addr]
	in	al, dx
	inc	al
	jnz	@F
	mov	[prog_err], PROG_ERR_NO_RSA_98II
	jmp	program_end
	mov	ax, dx			; ax=base addr
	call	set_16550_io_ports
@@:	mov	al, [specified_int]
  .if(al == 4)
	mov	al, 41h
  .endif
	mov	[int_serial], al
.elseif(al != COM_LINE_1st)
	mov	dl, al			; save
	; 9861K チェック
	mov	ax, [specified_addr]
	mov	byte ptr [io_offset_9861k + 1], al	; bit11 ～ bit8
	call	get_9861k_int
	jnc	@F
	mov	[prog_err], PROG_ERR_NO_9861K
	jmp	program_end
@@:	mov	[int_serial], al
	mov	ax, [io_offset_9861k]
	cmp	dl, COM_LINE_9861K_2nd
	jnz	@F
	mov	dx, 0b1h
	add	dx, ax
	mov	[serial_io_dat], dx
	mov	dx, 0b3h
	add	dx, ax
	mov	[serial_io_cmd], dx
	mov	dx, 0b0h
	add	dx, ax
	mov	[serial_io_sys], dx
	jmp	L1
@@:	mov	dx, 0b9h
	add	dx, ax
	mov	[serial_io_dat], dx
	mov	dx, 0bbh
	add	dx, ax
	mov	[serial_io_cmd], dx
	mov	dx, 0b2h
	add	dx, ax
	mov	[serial_io_sys], dx
.endif
L1:	;--- INT -> IRQ
	mov	al, [int_serial]
	cmp	al, 0ffh
	jz	@F
	call	int_to_irq
	mov	[irq_serial], al
	cmp	al, 7			; RS-232C の EOI 送信関数設定
	jbe	@F
	mov	[serial_eoi_func], offset send_EOI_to_slave
@@:	mov	al, [int_mouse]
	call	int_to_irq
	mov	[irq_mouse], al
	;--- 使う IRQ がもともとマスクされていたか保存
	mov	cl, [irq_serial]
	call	check_imr_mask
	jnc	@F
	or	[orig_imr_mask], IMR_FLAG_SERIAL
@@:	mov	cl, [irq_mouse]
	call	check_imr_mask
	jnc	@F
	or	[orig_imr_mask], IMR_FLAG_MOUSE
@@:	;--- RS-232C セットアップ
.if([com_line] >= COM_LINE_2ndCCU)
	;--- 16550 互換回線セットアップ
	; DSR を読んで相手がいるかチェック
	mov	dx, [second_ccu_io_msr]
	in	al, dx
	test	al, 20h
	jnz	@F
	FAILED_TO_SETUP_TABLET msg_err_no_tablet
@@:	; 11650 の割込マスク
	mov	dx, [second_ccu_io_lcr]
	in	al, dx
	mov	ah, al				; save
	and	al, 7fh
	out	dx, al
	xor	al, al
	mov	dx, [second_ccu_io_ier]
	out	dx, al
	; FIFO 無効
	mov	dx, [second_ccu_io_fcr]
	out	dx, al
	; FIFO 設定(割込1バイト, DMAモード0, FIFO 許可)
;	mov	al, 00000111b
;	mov	dx, [second_ccu_io_fcr]
;	out	dx, al
	; 分周比設定
	mov	al, ah
	or	al, 80h
	mov	dx, [second_ccu_io_lcr]
	out	dx, al				; DLAB=1
	xor	al, al
	mov	dx, [second_ccu_io_dh]
	out	dx, al
	mov	al, 12
	mov	dx, [second_ccu_io_dl]
	out	dx, al
	; パリティ, ストップビット長, データビット長設定(DLAB=0)
	mov	al, 0000011b
	mov	dx, [second_ccu_io_lcr]
	out	dx, al
.else
	;--- 標準 RS-232C セットアップ
	; DSR を読んで相手がいるかチェックする
	mov	dx, [serial_io_cmd]
	in	al, dx
	test	al, 80h
	jnz	@F
	FAILED_TO_SETUP_TABLET msg_err_no_tablet
@@:	;--- 8253A のカウンタ2を 9600bps 用に設定する
	xor	ax, ax
	mov	es, ax
	mov	al, 10110110b ; カウンタ2, 16bit読書, 方形波, バイナリカウント
	out	77h, al
	WAIT_5FH 2
.if(byte ptr es:[0501h] & 80h)
	mov	ax, 13			; 8MHz 系
.else
	mov	ax, 16			; 5MHz 系
.endif
	out	75h, al
	WAIT_5FH 2
	mov	al, ah
	out	75h, al
	mov	ax, cs
	mov	es, ax
	; RS-232C 9600bps 設定
	mov	dx, [serial_io_cmd]
	xor	al, al
REPEAT 3
	out	dx, al		; dummy instruction
	WAIT_5FH 6
ENDM
	mov	al, 40h
	out	dx, al		; software reset
	WAIT_5FH 6
	mov	al, 01001110b ; 1 stop bit, non parity, 8bit, 1/16 clock
	out	dx, al
.endif
	;--- RS-232C の初期設定だけ先にやってタブレットと通信する
	; RS-232C 割込ベクタ保存
	mov	ah, 35h
	mov	al, [irq_serial]
	add	al, 08h
	int	21h
	mov	[orig_vec_serial + 0], bx
	mov	[orig_vec_serial + 2], es
	; RS-232C 割込ベクタ設定
	mov	ah, 25h			; al は変化していない筈
	mov	dx, offset serial_int_hander
	int	21h
	mov	ax, cs
	mov	es, ax
	; RS-232C 割込マスク解除
	mov	cl, [irq_serial]
	call	clear_imr_mask
	;--- タブレットモデル名チェック
	; serial_wptr, rptr, cnt, err, total は初期値そのままで OK
	mov	[serial_mode], SER_MODE_GET_MODEL
	; RS-232C 設定(RTS=off であることに注意)
.if([com_line] >= COM_LINE_2ndCCU)
	mov	al, 01h
	mov	dx, [second_ccu_io_mcr]
	out	dx, al				; RTS=off
	mov	dx, [second_ccu_io_ier]
	out	dx, al			; 受信割込許可
.else
	mov	dx, [serial_io_cmd]
	mov	al, 00010111b
	out	dx, al
	WAIT_5FH 6
	; システムポートの受信割込マスク解除
	mov	dx, [serial_io_sys]
  .if([com_line] == COM_LINE_1st)
	in	al, dx
	and	al, 0f8h
	or	al, 01h
  .else
	mov	al, 01h		; 9861K の場合マスクを in で取得できない
  .endif
	out	dx, al
	WAIT_5FH 2
.endif
	; リセットとモデル名取得命令を送る
	mov	bx, offset ser_data_init_and_get_model
	mov	cx, 5 + 6
	call	send_serial
	; モデル名取得ループ
	mov	cx, 500		; 500[ms]位待つ
wait_model_loop:
	call	wait_1ms
	test	[serial_flags], SER_FLAG_READ_OK
	jnz	model_get_ok
	cmp	[serial_err], SER_NO_ERR	; エラーチェック
	jnz	@F
	cmp	[serial_total], TABLET_NAME_BUF_LEN ; バッファ長程度待つ
	jae	@F
	loop	wait_model_loop
@@:	; モデル名取得失敗
	FAILED_TO_SETUP_TABLET msg_err_no_tablet
	; モデル名取得成功 (タブレット名を表示)
model_get_ok:
	call	check_tablet_name		; モデル名確認
	jnc	@F
	FAILED_TO_SETUP_TABLET msg_err_not_supported
@@:	;--- タブレットが返す座標範囲取得
	xor	ax, ax
	mov	[serial_wptr], offset serial_buf
	mov	[serial_rptr], offset serial_buf
	mov	[serial_cnt], ax
	mov	[serial_total], ax
	mov	[serial_err], SER_NO_ERR
	mov	[serial_mode], SER_MODE_GET_AREA
	and	[serial_flags], NOT SER_FLAG_READ_OK
	; 座標範囲取得命令送信
.if([tablet_model] == TABLET_MODEL_XD)
	mov	bx, offset ser_data_get_tablet_area_xd
	mov	cx, 6 + 6
.else
	mov	bx, offset ser_data_get_tablet_area
	mov	cx, 6
.endif
	call	send_serial
	; 座標範囲取得ループ
	mov	cx, 500
wait_area_loop:
	call	wait_1ms
	test	[serial_flags], SER_FLAG_READ_OK
	jnz	area_get_ok
	cmp	[serial_err], SER_NO_ERR
	jnz	@F
	cmp	[serial_total], TABLET_AREA_STR_BUF_LEN
	jae	@F
	loop	wait_area_loop
@@:	; 座標範囲取得失敗
	FAILED_TO_SETUP_TABLET msg_err_failed_to_get_prop
	; 座標範囲取得成功
area_get_ok:
	call	parse_tablet_area ; 座標範囲パース
	jnc	@F
	FAILED_TO_SETUP_TABLET msg_err_failed_to_get_prop
@@:	; 余白幅を考慮した座標範囲を計算
	; 余白幅の合計が 65535 以上か、タブレットの座標範囲以上ならエラー
	mov	dx, [tablet_margin_l]
	mov	cx, [tablet_margin_t]
	add	dx, [tablet_margin_r]
	jc	invalid_margin
	add	cx, [tablet_margin_b]
	jc	invalid_margin
	mov	ax, [tablet_area_width]
	mov	bx, [tablet_area_height]
	cmp	dx, ax
	jae	invalid_margin
	cmp	cx, bx
	jb	@F
invalid_margin:
	mov	[prog_err], PROG_ERR_INVALID_MARGIN
	jmp	failed_on_tablet2
@@:	sub	ax, dx
	sub	bx, cx
	mov	[tablet_real_width], ax
	mov	[tablet_real_height], bx
	; タブレットデータ受信準備(RTS=on)
.if([com_line] >= COM_LINE_2ndCCU)
	mov	al, 03h
	mov	dx, [second_ccu_io_mcr]
	out	dx, al
.else
	mov	dx, [serial_io_cmd]
	mov	al, 00110111b
	out	dx, al
	WAIT_5FH 6
.endif
	xor	ax, ax
	mov	[serial_wptr], offset serial_buf
	mov	[serial_rptr], offset serial_buf
	mov	[serial_cnt], ax
	mov	[serial_total], ax
	mov	[serial_err], SER_NO_ERR
	and	[serial_flags], NOT SER_FLAG_READ_OK
	mov	[serial_mode], SER_MODE_PARSE_DATA
	;--- マウス関係セットアップ
	; マウス割り込み停止及びマスク
	mov	dx, MOUSE_IO_PORT_C
	mov	al, 10h			; 割込不許可(INT#=1)
	out	dx, al
	WAIT_5FH 2
	mov	cl, [irq_mouse]
	call	set_imr_mask
	; マウス初期設定
	mov	dx, MOUSE_IO_CTRL_REG
	mov	al, MOUSE_IO_CTRL_VAL
	out	dx, al			; 8255A 初期設定
	WAIT_5FH 2
	mov	dx, MOUSE_IO_INT_FREQ
	mov	al, [mouse_int_freq]
	out	dx, al			; マウス割込頻度設定
	; マウス割り込みベクタ保存
	mov	ah, 35h
	mov	al, [irq_mouse]
	add	al, 08h
	int	21h
	mov	[orig_vec_mouse + 0], bx
	mov	[orig_vec_mouse + 2], es
	; マウス割込ベクタ設定
	mov	ah, 25h
	mov	dx, offset mouse_int_hander
	int	21h
	mov	ax, cs
	mov	es, ax
	; マウス割込マスク解除及び割込許可
	mov	cl, [irq_mouse]
	call	clear_imr_mask
	mov	dx, MOUSE_IO_PORT_C
	xor	al, al			; HC=0, 割込許可 (INT#=0)
	out	dx, al
	WAIT_5FH 2
	;--- 制御ベクタセットアップ
	; 保存
	mov	ax, 3500h + CTRL_VECTOR
	int	21h
	mov	[orig_vec_ctrl + 0], bx
	mov	[orig_vec_ctrl + 2], es
	; 設定
	mov	ax, 2500h + CTRL_VECTOR
	mov	dx, offset ctrl_int_handler
	int	21h
	;--- 常駐前処理(残り諸々)
	mov	ax, ds:[2ch]
	mov	es, ax
	mov	ah, 49h
	int	21h				; 環境変数領域解放
	mov	ax, cs
	mov	es, ax
	call	print_tablet_info
	call	print_driver_settings
	mov	ax, offset msg_ok_stay
	call	putstr_newline
	or	[prog_flags], PROG_FLAG_DO_STAY
	jmp	program_end
	;--- タブレット関係の失敗 (ds:si=エラーメッセージ)
	; RS-232C の後始末しかしないため注意
failed_on_tablet:
	mov	[prog_err], PROG_ERR_TABLET_SETUP
failed_on_tablet2:			; [prog_err] を別の値にしたい場合用
	mov	al, [com_line]
.if(al >= COM_LINE_2ndCCU)
	xor	al, al
	mov	dx, [second_ccu_io_ier]
	out	dx, al
.else
	; システムポートで RS-232C 割込マスク
	mov	dx, [serial_io_sys]
  .if(al == COM_LINE_1st)
	in	al, dx
	and	al, 0f8h
  .else
	xor	al, al
  .endif
	out	dx, al
.endif
	; 割込ベクタと IMR 復帰 (マウスと制御ベクタはこの時点では未設定)
	call	restore_serial_vec
	call	restore_imr
	jmp	program_end
do_stay_j ENDP

;---------------------------------------
; 取得したタブレット名がサポートしている型番か調べる
; 引数: 無し
; 戻値: CF=結果(0=サポートしている, 1=サポートしていない)
; 破壊: ax, dl
check_tablet_name PROC
	cmp	word ptr [tablet_name], '#~'
	jnz	ng
	mov	ax, word ptr [tablet_name + 2]
	xor	dl, dl
	cmp	ax, 'DU'	; UD series
	je	ok_ud
	cmp	ax, 'TK'	; KT series
	je	ok_kt
	cmp	ax, 'DG'	; GD series
	je	ok_gd
	cmp	ax, 'DX'	; XD series
	je	ok_xd
ng:	stc
	ret
ok_xd:	inc	dl	; XD: dl == 3
ok_gd:	inc	dl	; GD: dl == 2
ok_kt:	inc	dl	; KT: dl == 1
ok_ud:			; UD: dl == 0
.if(dl <= 1)
	mov	[tablet_proto], PROTO_W_IV
	mov	[tablet_packet_len], PROTO_W_IV_PACKET_LEN
	test	dl, dl
	jnz	@F
	mov	[tablet_model], TABLET_MODEL_UD
	mov	ax, PRESSURE_THRESHOLD_UD
	jmp	L1
@@:	mov	[tablet_model], TABLET_MODEL_KT
	mov	ax, PRESSURE_THRESHOLD_KT
.else
	mov	[tablet_proto], PROTO_W_V
	mov	[tablet_packet_len], PROTO_W_V_PACKET_LEN
	cmp	dl, 3
	jz	@F
	mov	[tablet_model], TABLET_MODEL_GD
	mov	ax, PRESSURE_THRESHOLD_GD
	jmp	L1
@@:	mov	[tablet_model], TABLET_MODEL_XD
	mov	ax, PRESSURE_THRESHOLD_XD
.endif
L1:	test	[specified_opts], OPT_P_THRESHOLD
	jnz	ok
	mov	[tablet_p_threshold], ax
ok:	clc
	ret
check_tablet_name ENDP

;---------------------------------------
; 取得した座標範囲の値をパースする
; 引数: 無し
; 戻値: CF=結果(0=パース成功, 1=パース失敗)
parse_tablet_area PROC
	cmp	word ptr [tablet_area_str], 'C~'
	jnz	ng
	; X 方向パース
	mov	bx, offset tablet_area_str + 2
	call	parse_dec_str
	jc	ng
	mov	[tablet_area_width], ax
	; 区切りのカンマ確認
	add	bx, si
	cmp	byte ptr [bx], ','
	jnz	ng
	; Y 方向パース
	inc	bx
	call	parse_dec_str
	jc	ng
	mov	[tablet_area_height], ax
	; 終端の NULL 文字確認
	cmp	byte ptr [bx + si], 0
	jnz	ng
	clc
	ret
ng:	stc
	ret
parse_tablet_area ENDP

;---------------------------------------
; 認識したタブレットの情報を表示する
; 引数, 戻値: 無し
print_tablet_info PROC
	;--- 9861K 関連情報
	mov	al, [com_line]
.if((al == COM_LINE_9861K_2nd) || (al == COM_LINE_9861K_3rd))
	mov	ax, offset msg_header_9861k_0
	call	putstr
	mov	ax, [serial_io_sys]
	call	print_ax_hex
	mov	ax, offset msg_header_9861k_1
	call	putstr
	mov	al, [int_serial]
	call	print_int
	mov	ax, offset msg_newline
	call	putstr
.endif
	;--- タブレット型番
	mov	ax, offset msg_header_tablet_name
	call	putstr
	mov	ax, offset tablet_name + 2 ; '~#' なしで表示
	call	putstr_newline
	;--- タブレット座標範囲
	mov	ax, offset msg_header_tablet_area
	call	putstr
	mov	ax, [tablet_area_width]
	call	print_ax_dec
	mov	al, ','
	int	29h
	mov	ax, [tablet_area_height]
	call	print_ax_dec
	mov	al, ')'
	int	29h
	mov	ax, offset msg_newline
	call	putstr
	;--- 筆圧閾値
	mov	ax, offset msg_header_p_threshold
	call	putstr
	mov	ax, [tablet_p_threshold]
	call	print_ax_dec
	mov	al, ','
	int	29h
	mov	al, ' '
	int	29h
	;--- 余白幅
	mov	ax, offset msg_header_margin
	call	putstr
	mov	ax, [tablet_margin_t]
	call	print_ax_dec
	mov	al, ','
	int	29h
	mov	ax, [tablet_margin_b]
	call	print_ax_dec
	mov	al, ','
	int	29h
	mov	ax, [tablet_margin_l]
	call	print_ax_dec
	mov	al, ','
	int	29h
	mov	ax, [tablet_margin_r]
	call	print_ax_dec
	mov	al, ')'
	int	29h
	mov	ax, offset msg_newline
	call	putstr
	ret
print_tablet_info ENDP

;---------------------------------------
; ドライバの設定内容を表示する
; 引数, 戻値: 無し
print_driver_settings PROC
	push	bx
	mov	bx, offset msg_header_drv_flags
	mov	al, [driver_flags_s]
.if(!(al & DRIVER_FLAG_S_SYNC_CURSOR))
	mov	word ptr [bx + 14], 0b396h	; 「有」->「無」
.endif
.if(!(al & DRIVER_FLAG_S_EN_DBG_FUNC))
	mov	word ptr [bx + 40], 0b396h
.endif
	mov	ax, bx
	call	putstr_newline
	pop	bx
	ret
print_driver_settings ENDP


;---------------------------------------
; 常駐解除処理
do_remove_j PROC
	;--- 常駐チェック
	call	check_stay		; ax=常駐セグメントアドレス
	jc	@F
	mov	[prog_err], PROG_ERR_NOT_STAYED
	jmp	program_end
@@:	;--- 常駐部分のデータをコピー
	mov	bp, ax
	call	fetch_tsr_data
	;--- RS-232C 後始末
	; RS-232C の割込を全部マスク
	mov	al, [com_line]
.if(al >= COM_LINE_2ndCCU)
	xor	al, al
	mov	dx, [second_ccu_io_ier]
	out	dx, al
.else
	mov	dx, [serial_io_sys]
  .if(al == COM_LINE_1st)
	in	al, dx
	and	al, 0f8h
  .else
	xor	al, al
  .endif
	out	dx, al
.endif
	; マウス割り込み停止
	mov	dx, MOUSE_IO_PORT_C
	mov	al, 10h
	out	dx, al
	; IMR で RS-232C とマウスの割込をマスク
	mov	cl, [irq_serial]
	call	set_imr_mask
	mov	cl, [irq_mouse]
	call	set_imr_mask
	;--- 割込ベクタと IMR 復帰
	call	restore_serial_vec
	call	restore_mouse_vec
	call	restore_ctrl_vec
	call	restore_imr
	;--- 常駐メモリ解放
	mov	es, bp
	mov	ah, 49h
	int	21h
	mov	ax, cs
	mov	es, ax
	;---
	mov	ax, offset msg_ok_remove
	call	putstr_newline
	jmp	program_end
do_remove_j ENDP


;---------------------------------------
; TSR からデータをコピーする
; 引数: ax=常駐セグメントアドレス
; 戻値: 無し
fetch_tsr_data PROC
	push	ds
	push	es
	push	di
	push	si
	push	cx
	;---
	mov	di, ds
	mov	es, di
	mov	ds, ax
	mov	di, offset tsr_data_head
	mov	si, di
	mov	cx, (offset tsr_tail - offset tsr_data_head) / 2 + 1
	rep	movsw
	;---
	pop	cx
	pop	si
	pop	di
	pop	es
	pop	ds
	ret
fetch_tsr_data ENDP


;---------------------------------------
; 常駐チェック
; 引数: 無し
; 戻値: CF=結果(1=常駐している, 0=常駐していない)
;       ax=常駐している場合は常駐セグメント
check_stay PROC
	push	es
	push	bx
	;---
	mov	ax, 3500h + CTRL_VECTOR
	int	21h		; es:bx=interrupt vector
	;--- チェック用データ判定
	cmp	word ptr es:[bx + 3], 'MW'
	jnz	not_found
	cmp	word ptr es:[bx + 5], 'RD'
	jnz	not_found
	cmp	word ptr es:[bx + 7], 'V' ; NULL 終端なので word ptr で見る
	jnz	not_found
	stc
	mov	ax, es
	;---
@@:	pop	bx
	pop	es
	ret
not_found:
	clc
	jmp	@B
check_stay ENDP


;---------------------------------------
; 割込ベクタ復帰
; 引数, 戻値: 無し
; orig_vec_ctrl, orig_vec_serial, orig_vec_mouse が設定されている必要あり
restore_ctrl_vec PROC
	push	bx
	mov	al, CTRL_VECTOR
	mov	bx, offset orig_vec_ctrl
	call	restore_vec
	pop	bx
	ret
restore_ctrl_vec ENDP
restore_serial_vec PROC
	push	bx
	mov	al, [irq_serial]
	add	al, 8
	mov	bx, offset orig_vec_serial
	call	restore_vec
	pop	bx
	ret
restore_serial_vec ENDP
restore_mouse_vec PROC
	push	bx
	mov	al, [irq_mouse]
	add	al, 8
	mov	bx, offset orig_vec_mouse
	call	restore_vec
	pop	bx
	ret
restore_mouse_vec ENDP
; 割り込みベクタ復帰処理本体
; 引数: al=ベクタ番号, ds:bx=ベクタデータ保存位置
; 戻値: 無し
restore_vec PROC
	push	dx
	push	ds
	mov	dx, [bx]
	mov	ds, [bx + 2]
	mov	ah, 25h
	int	21h
	pop	ds
	pop	dx
	ret
restore_vec ENDP

;---------------------------------------
; プログラム引数のパース
; 引数: 無し
; 戻値: CF=成功?(0: 成功, 1: 失敗)
; 破壊: 全汎用レジスタ
;--- 失敗させるマクロ
FAIL_ARG_PARSE MACRO msg:REQ
	mov	[err_add_msg], offset msg
	jmp	fail
ENDM
;--- 文字列長チェックマクロ
CHECK_ARG_LEN MACRO msg:REQ
.if(bp <= bx)
	FAIL_ARG_PARSE msg
.endif
ENDM
;--- 空白チェックマクロ
CHECK_WHITESPACE MACRO msg:REQ
	mov	al, [bx]
	call	is_whitespace
.if(carry?)
	FAIL_ARG_PARSE msg
.endif
ENDM
;--- 終端チェックマクロ
CHECK_WHITESPACE_NULL MACRO msg:REQ
	mov	al, [bx]
	call	is_whitespace
.if(carry? && al != 0)		; 空文字でも NULL 文字でもないならエラー
	FAIL_ARG_PARSE msg
.endif
ENDM
parse_args PROC
	mov	bx, 81h
	mov	bp, [bx - 1]
	and	bp, 0ffh
	add	bp, bx			; bp=プログラム引数終端位置
	mov	byte ptr ds:[bp], 0	; NULL 終端にする
	;--- command 取得
	call	skip_whitespace
	CHECK_ARG_LEN msg_err_no_cmd
	mov	al, [bx]		; コマンド文字
	; 引数のないコマンド
	cmp	al, 'h'
	jnz	@F
	mov	[prog_mode], PROG_MODE_HELP
	jmp	L001
@@:	cmp	al, 'r'
	jnz	@F
	mov	[prog_mode], PROG_MODE_REMOVE
L001:	inc	bx
	CHECK_WHITESPACE_NULL msg_err_invalid_cmd
	clc
	ret
@@:	; 引数のあるコマンド
	cmp	al, 's'
	jnz	@F
	mov	[prog_mode], PROG_MODE_STAY
	jmp	L002
@@:	; 不明なコマンド
	FAIL_ARG_PARSE msg_err_invalid_cmd
L002:	inc	bx
	cmp	bx, bp
	jae	parse_opt_end
	CHECK_WHITESPACE msg_err_invalid_cmd
	;--- オプションパース処理
parse_options:
	call	skip_whitespace
	cmp	bx, bp
	jae	parse_opt_end
	mov	al, [bx]
.if((al != '/') && (al != '-'))
	FAIL_ARG_PARSE msg_err_invalid_opt
.endif
	inc	bx
	CHECK_ARG_LEN msg_err_invalid_opt
	mov	al, [bx]
	inc	bx
	call	parse_option_main
	jnc	parse_options
	FAIL_ARG_PARSE msg_err_invalid_opt
parse_opt_end:
	clc
	ret
	;--- 失敗処理
fail:	mov	[prog_err], PROG_ERR_ARG
	stc
	ret
;--- オプションパース処理
; 引数: al=オプション文字, bx=オプションに続く文字列, bp=引数文字列終端位置
; 戻値: CF=成功?(0: 成功, 1: 失敗), bx=オプションの次の文字列
parse_option_main:
	cmp	al, 'p'
	jz	L1
	cmp	al, 'c'
	jz	L1
	cmp	al, 'i'
	jz	L1
	cmp	al, 'k'
	jz	L1
	cmp	al, 's'
	jnz	Ld			; 'p', 'c', 'i', 'k', 's' 以外
	; 10進数を1個パースすればいいオプションは共通で扱う
L1:	mov	cl, al			; save
	call	skip_whitespace
	CHECK_ARG_LEN msg_err_invalid_opt
	call	parse_dec_str
	jc	invalid_opt
.if(cl == 'p')
	mov	[tablet_p_threshold], ax
	or	[specified_opts], OPT_P_THRESHOLD
.elseif(cl == 'c')
	cmp	ax, COM_LINE_MAX
	ja	invalid_opt
	mov	[com_line], al
	or	[specified_opts], OPT_COM_LINE
.elseif(cl == 'i')
	cmp	ax, 6			; INT0～INT6
	ja	invalid_opt
	mov	[specified_int], al
	or	[specified_opts], OPT_INT
.elseif(cl == 's')
	test	ax, ax
	jnz	@F
	and	[driver_flags_s], NOT DRIVER_FLAG_S_SYNC_CURSOR
@@:	or	[specified_opts], OPT_SYNC_CURSOR
.else	; 'k'
	mov	[mky_dot_rate_specify], ax
	mov	[mky_dot_rate_x], ax
	mov	[mky_dot_rate_y], ax
	or	[specified_opts], OPT_MKY
.endif
	add	bx, si
	jmp	opt_ok
Ld:	;--- 'd'
	cmp	al, 'd'
	jnz	Lm
	or	[specified_opts], OPT_DEBUG_FUNC
	or	[driver_flags_s], DRIVER_FLAG_S_EN_DBG_FUNC
	jmp	opt_ok
Lm:	;--- 'm'
	cmp	al, 'm'
	jnz	La
	xor	cl, cl
	mov	di, offset tablet_margin_t
Lm0:	call	skip_whitespace
	CHECK_ARG_LEN msg_err_invalid_opt
	call	parse_dec_str
	jc	invalid_opt
	cmp	cl, 4
	jae	invalid_opt	; 5回目のパースに成功したらエラー
	inc	cl
	mov	[di], ax
	add	bx, si
	call	skip_whitespace
	cmp	byte ptr [bx], ','
	jnz	Lm1
	inc	bx
	add	di, 2
	jmp	Lm0
Lm1:	test	cl, cl
	jz	invalid_opt		; cl=0 ならエラー
	cmp	cl, 3
	je	invalid_opt		; cl=3 ならエラー
	cmp	cl, 2			; cl=1,2,4 の何れか
	je	Lm2
	ja	Lm3
	mov	ax, [tablet_margin_t]	; 1個なら全部同じ幅
	mov	[tablet_margin_b], ax
	mov	[tablet_margin_l], ax
	mov	[tablet_margin_r], ax
	jmp	Lm3
Lm2:	mov	ax, [tablet_margin_b]	; 2個なら上下と左右の指定
	mov	dx, [tablet_margin_t]
	mov	[tablet_margin_b], dx
	mov	[tablet_margin_l], ax
	mov	[tablet_margin_r], ax
Lm3:	or	[specified_opts], OPT_TABLET_MARGIN
	jmp	opt_ok
La:	;--- 'a'
	cmp	al, 'a'
	jnz	invalid_opt
	xor	dx, dx
	xor	cl, cl
	call	skip_whitespace
	CHECK_ARG_LEN msg_err_invalid_opt
La0:	mov	al, [bx]
.if(cl)
	test	al, al		; '\0'
	jz	La1
	call	is_whitespace
	jnc	La1
.endif
	call	parse_hex_char
	jc	invalid_opt
	SHFT_L	dx, 4
	or	dl, al
	inc	cl
	cmp	cl, 5
	jae	invalid_opt	; 最大4桁
	inc	bx
	jmp	La0
La1:	test	cl, cl
	jz	invalid_opt	; 1文字も数値をパースしていない
	mov	[specified_addr], dx
	or	[specified_opts], OPT_IO_ADDR
	;---
opt_ok:	clc
	ret
invalid_opt:
	stc
	ret
parse_args ENDP


;---------------------------------------
; ヘルプ表示
do_help_j PROC
	mov	ax, offset msg_usage
	call	putstr
	jmp	program_end
do_help_j ENDP


;-----------------------------------------------------------
; RS-232C でデータを送る
; 引数: ds:bx=送信データ, cx=バイト数
; 戻値: 無し
; jmp 先の関数で ret する
send_serial PROC
	cmp	[com_line], COM_LINE_2ndCCU
	jb	send_serial_8251
	jmp	send_serial_2ndCCU
send_serial ENDP
;---------------------------------------
; 8251 でデータを送る(TxRDY をポーリング)
; 引数: ds:bx=送信データ, cx=バイト数
; 戻値: 無し
send_serial_8251 PROC
	push	bx
	push	cx
	push	dx
	;---
	add	cx, bx
L1:	cmp	bx, cx
	jae	send_end
	;--- TxRDY=1 を待つ
	mov	dx, [serial_io_cmd]
@@:	in	al, dx			; status read
	test	al, 1
	jz	@B
	;--- データ送信
	mov	dx, [serial_io_dat]
	mov	al, [bx]
	out	dx, al
	inc	bx
	jmp	L1
send_end:
	;---
	pop	dx
	pop	cx
	pop	bx
	ret
send_serial_8251 ENDP
;---------------------------------------
; 2ndCCU でデータを送る(TxRDY をポーリング)
; 引数: ds:bx=送信データ, cx=バイト数
; 戻値: 無し
send_serial_2ndCCU PROC
	push	bx
	push	cx
	push	dx
	;---
	add	cx, bx
L1:	cmp	bx, cx
	jae	send_end
	;--- TEMT を待つ
	mov	dx, [second_ccu_io_lsr]
@@:	in	al, dx
	test	al, 40h
	jz	@B
	;--- データ送信
	mov	al, [bx]
	mov	dx, [second_ccu_io_data]
	out	dx, al
	inc	bx
	jmp	L1
send_end:
	;---
	pop	dx
	pop	cx
	pop	bx
	ret
send_serial_2ndCCU ENDP


;---------------------------------------
; INT -> IRQ 変換
; 引数: al=INT(BCD)
; 戻値: al=IRQ
int_to_irq PROC
.if(al == 41h)
	mov	al, 10
.elseif(al == 42h)
	mov	al, 11
.else
	push	bx
	xor	bh, bh
	mov	bl, al
	mov	al, [bx + tbl]
	pop	bx
.endif
	ret
tbl db 3, 5, 6, 9, ?, 12, 13	; ?=4
int_to_irq ENDP

;---------------------------------------
; IMR を初期状態に戻す
; 引数, 戻値: 無し
; irq_serial, irq_mouse, orig_imr_mask が適切に設定されている必要がある
restore_imr PROC
	push	cx
	mov	ch, [orig_imr_mask]
	mov	cl, [irq_serial]
.if(ch & IMR_FLAG_SERIAL)
	call	set_imr_mask
.else
	call	clear_imr_mask
.endif
	mov	cl, [irq_mouse]
.if(ch & IMR_FLAG_MOUSE)
	call	set_imr_mask
.else
	call	clear_imr_mask
.endif
	pop	cx
	ret
restore_imr ENDP

;---------------------------------------
; IMR にマスクを設定
; 引数: cl=マスク設定する IRQ
; 戻値: 無し
set_imr_mask PROC
	push	dx
	mov	ah, 1
.if(cl <= 7)
	shl	ah, cl
	mov	dx, 02h
.else
	sub	cl, 8
	shl	ah, cl
	add	cl, 8
	mov	dx, 0ah
.endif
	cli
	in	al, dx
	or	al, ah
	out	dx, al
	sti
	pop	dx
	ret
set_imr_mask ENDP

;---------------------------------------
; IMR にマスクが設定されているか判定
; 引数: cl=判定対象の IRQ
; 戻値: CF=結果(0: 非マスク, 1: マスク中)
check_imr_mask PROC
	push	dx
	mov	ah, 1
.if(cl <= 7)
	shl	ah, cl
	mov	dx, 02h
.else
	sub	cl, 8
	shl	ah, cl
	add	cl, 8
	mov	dx, 0ah
.endif
	in	al, dx
	pop	dx
	test	al, ah
	jnz	@F
	clc
	ret
@@:	stc
	ret
check_imr_mask ENDP

;---------------------------------------
; PC-9861K 互換の RS-232C が使う INT を取得する
; 引数: [com_line]=通信回線(COM_LINE_9861K_* 以外の時は呼出禁止)
;       [io_offset_9861k]=9861K IO アドレスオフセット(PC-9801-101 で設定可)
; 戻値: al=INT(BCD), CF=成功?(0=成功, 1=失敗)
get_9861k_int PROC
	push	bx
	push	dx
	;---
	mov	dx, [io_offset_9861k]
	cmp	[com_line], COM_LINE_9861K_3rd
	je	L3rd
	add	dx, 0b0h
	mov	bx, offset tbl_2nd
	jmp	@F
L3rd:	add	dx, 0b2h
	mov	bx, offset tbl_3rd
@@:	;---
	in	al, dx
	cmp	al, 0ffh
	je	ng			; ボード見つからず
	and	al, 03h
	xor	ah, ah
	add	bx, ax
	mov	al, [bx]
	;---
	clc
@@:	pop	dx
	pop	bx
	ret
ng:	stc
	jmp	@B
	;---
get_9861k_int ENDP
tbl_2nd db 00h, 01h, 02h, 03h
tbl_3rd db 00h, 41h, 05h, 06h		; INT 41 固定

;---------------------------------------
; 16550 にアクセスするための IO アドレスを設定する
; 引数: ax=ベースアドレス
; 戻値: 無し
set_16550_io_ports PROC
	push	bx
	;---
	mov	bx, offset second_ccu_io_data
	mov	[bx + 0], ax	; DATA (base + 0)
	mov	[bx + 2], ax	; DL (base + 0)
	inc	ax
	mov	[bx + 4], ax	; IER (base + 1)
	mov	[bx + 6], ax	; DH (base + 1)
	inc	ax
	mov	[bx + 8], ax	; IIR (base + 2)
	mov	[bx + 10], ax	; FCR (base + 2)
	inc	ax
	mov	[bx + 12], ax	; LCR (base + 3)
	inc	ax
	mov	[bx + 14], ax	; MCR (base + 4)
	inc	ax
	mov	[bx + 16], ax	; LSR (base + 5)
	inc	ax
	mov	[bx + 18], ax	; MSR (base + 6)
	inc	ax
	mov	[bx + 20], ax	; SCR (base + 7)
	;---
	pop	bx
	ret
set_16550_io_ports ENDP


;---------------------------------------
; out 5fh を使って約 1ms 待つ
; 引数, 戻値: 無し
wait_1ms PROC
	push	cx
	mov	cx, 1667	; ceil(1000[us] / 0.6[us])
@@:	out	5fh, al
	loop	@B
	pop	cx
	ret
wait_1ms ENDP


;---------------------------------------
; 空白文字か判定する
; 引数: al=判定対象文字
; 戻値: CF=結果(0: 空白文字である, 1: 空白文字でない)
is_whitespace PROC
	cmp	al, ' '
	je	@F
	cmp	al, 09h		; 水平タブ
	je	@F
	cmp	al, 0bh		; 垂直タブ
	je	@F
	cmp	al, 0dh
	je	@F
	cmp	al, 0ah
	je	@F
	stc
	ret
@@:	clc
	ret
is_whitespace ENDP

;---------------------------------------
; 空白文字以外の文字が出るまで bx を進める
; 引数: ds:bx=処理対象文字列
; 戻値: bx=空白でない文字の位置
skip_whitespace PROC
L1:	mov	al, [bx]
	call	is_whitespace
	jnc	@F
	ret
@@:	inc	bx
	jmp	L1
skip_whitespace ENDP

;---------------------------------------
; NULL 終端文字列表示
; 引数: ds:ax=文字列の先頭アドレス
; 戻値: 無し
putstr PROC PRIVATE
	push	si
	mov	si, ax
L1:	mov	al, [si]
	test	al, al
	jz	@F
	int	29h
	inc	si
	jmp	L1
@@:	pop	si
	ret
putstr ENDP

;---------------------------------------
; NULL 終端文字列と改行表示
; 引数: ds:ax=文字列の先頭アドレス
; 戻値: 無し
putstr_newline PROC PRIVATE
	call	putstr
	mov	al, 0dh
	int	29h
	mov	al, 0ah
	int	29h
	ret
putstr_newline ENDP

;---------------------------------------
; INT 表示 (0, 1, 2, 3, 41, 42, 5, 6 の何れかを表示)
; 引数: al=INT(bcd)
; 戻値: 無し
print_int PROC
	cmp	al, 41h
	jne	@F
	mov	ax, offset str_41
	jmp	putstr
@@:	cmp	al, 42h
	jne	@F
	mov	ax, offset str_42
	jmp	putstr
@@:	cmp	al, 6
	ja	@F
	add	al, '0'
	int	29h
@@:	ret
print_int ENDP
str_41	db '41', 0
str_42	db '42', 0

;---------------------------------------
; ax の値を16進数4桁で表示する
; 引数: ax=表示する値
; 戻値: 無し
print_ax_hex PROC
	push	ax
	mov	al, ah
	call	print_al_hex
	pop	ax
	;--- print_al_hex に流れ落ちる ---
print_ax_hex ENDP
;---------------------------------------
; al の値を16進数2桁で表示する
; 引数: al=表示する値
; 戻値: 無し
print_al_hex PROC
	mov	ah, al
	SHFT_R	al, 4
	and	ah, 0fh
	call	print_hex_single
	mov	al, ah
	call	print_hex_single
	ret
print_hex_single:
	cmp	al, 0ah
	jb	@F
	add	al, 'a' - 0ah
	jmp	L1
@@:	add	al, '0'
L1:	int	29h
	ret
print_al_hex ENDP

;---------------------------------------
; ax の値を正の10進数と思って表示する
; 引数: ax=表示する値
; 戻値: 無し
print_ax_dec PROC
	test	ax, ax
	jnz	@F
	mov	al, '0'
	int	29h
	ret
@@:	push	dx
	push	cx
	push	bx
	;--- 各桁の値を求める
	mov	cx, 10
	mov	bx, offset buf - 1
@@:	xor	dx, dx
	div	cx		; dx:ax / cx -> ax=商, dx=余
	inc	bx
	mov	[bx], dl
	cmp	bx, offset buf + 4
	jb	@B
	;--- 表示処理 (dl の LSB=1 で表示中)
	mov	dx, (5 SHL 8) ; (dh, dl) = (5, 0)
@@:	mov	al, [bx]
	cmp	al, 0
	jnz	print
	test	dl, 1
	jnz	print
	jmp	skip
print:	or	dl, 1
	add	al, '0'
	int	29h
skip:	dec	bx
	dec	dh
	jnz	@B
	;---
	pop	bx
	pop	cx
	pop	dx
	ret
buf	db 5 dup(?)		; 各桁の値の一時保存用
print_ax_dec ENDP

;---------------------------------------
; 正の10進数文字列(最大値:65535)をパースする
; 引数: ds:bx=パース対象の文字列
; 戻値: ax=パースした値, si=文字数, CF=成功?(0: 成功, 1: 失敗)
parse_dec_str PROC
	push	dx
	push	cx
	;---
	xor	si, si
	xor	ax, ax
	xor	ch, ch		; 0bit 目が1ならパース中
	;--- パース処理ループ
L1:	mov	cl, [bx + si]
	cmp	cl, '9'
	ja	exit
	cmp	cl, '0'
	jb	exit
	; 冒頭の連続する '0' は見逃す
	jnz	do_parse
	test	ch, 1
	jnz	do_parse
	inc	si
	jmp	L1
do_parse:
	; パース処理本体
	xor	ch, ch
	sub	cl, '0'		; cx=足す数
	mov	dx, 10
	mul	dx		; dx:ax=10*ax
	test	dx, dx
	jnz	too_large
	add	ax, cx
	jc	too_large
	inc	si
	or	ch, 1		; パース中フラグ立てる
	jmp	L1
	;---
exit:	clc
@@:	pop	cx
	pop	dx
	ret
too_large:
	stc
	jmp	@B
parse_dec_str ENDP

;---------------------------------------
; 16進数1文字をパースする
; 引数: al=対象文字
; 戻値: al=パースした値, CF=成功?(0=成功, 1=失敗)
parse_hex_char PROC
.if((al >= '0') && (al <= '9'))
	sub	al, '0'
	clc
.elseif((al >= 'a') && (al <= 'f'))
	add	al, 0ah - 'a'
	clc
.elseif((al >= 'A') && (al <= 'F'))
	sub	al, 0ah - 'A'
	clc
.else
	stc
.endif
	ret
parse_hex_char ENDP


;-----------------------------------------------------------
; stack
	even
	db 384 dup(?)
stack_start:

CODE ENDS
END start
